# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import os
import unittest

from byot import (
    features,
    scenarii,
)

from byov import (
    config,
    ssh,
)
from byov.tests import (
    features as vms_features,
    fixtures as vms_fixtures,
)
from byov.vms import lxd


load_tests = scenarii.load_tests_with_scenarios


class TestInfoFromPath(unittest.TestCase):

    scenarios = scenarii.multiply_scenarios(
        # key
        ([('rsa', dict(algo='rsa', key_name='rsa')),
          ('dsa', dict(algo='dsa', key_name='dsa')),
          ('ecdsa', dict(algo='ecdsa', key_name='ecdsa')),
          ('unknown', dict(algo=None, key_name='foo'))]),
        # visibility
        ([('private', dict(visible='private', suffix='')),
          ('public', dict(visible='public', suffix='.pub'))]))

    def assertInfoFromPath(self, expected, key_path):
        self.assertEqual(expected, ssh.infos_from_path(key_path))

    def test_key(self):
        self.assertInfoFromPath((self.algo, self.visible),
                                './{}{}'.format(self.key_name, self.suffix))


@features.requires(vms_features.ssh_feature)
class TestKeyGen(unittest.TestCase):

    scenarios = [('rsa', dict(algo='rsa', prefix='ssh-rsa ',
                              upper_type='RSA')),
                 ('dsa', dict(algo='dsa', prefix='ssh-dss ',
                              upper_type='DSA')),
                 ('ecdsa', dict(algo='ecdsa', prefix='ecdsa-sha2-nistp256 ',
                                upper_type='EC'))]

    def setUp(self):
        super(TestKeyGen, self).setUp()
        vms_fixtures.isolate_from_disk(self)

    def keygen(self, ssh_type, upper_type):
        private_path = os.path.join(self.uniq_dir, ssh_type)
        ssh.keygen(private_path, 'byov@test')
        self.assertTrue(os.path.exists(private_path))
        public_path = private_path + '.pub'
        self.assertTrue(os.path.exists(public_path))
        with open(public_path) as f:
            public = f.read()
        with open(private_path) as f:
            private = f.read()
        if vms_features.ssh_feature.version >= 'OpenSSH_7.9':
            comment_decoration = 'OPENSSH'
        else:
            comment_decoration = upper_type
        self.assertTrue(
            private.startswith('-----BEGIN {} PRIVATE KEY-----\n'.format(
                comment_decoration,)))
        self.assertTrue(
            private.endswith('-----END {} PRIVATE KEY-----\n'.format(
                             comment_decoration,)))
        self.assertTrue(public.endswith(' byov@test\n'))
        return private, public

    def test_key(self):
        private, public = self.keygen(self.algo, self.upper_type)
        self.assertTrue(public.startswith(self.prefix))


class TestSsh(unittest.TestCase):

    def setUp(self):
        super(TestSsh, self).setUp()
        vms_fixtures.setup_tests_config(self)
        vms_features.requires_existing_vm(self, 'byov-tests-lxd-debian')
        vms_fixtures.set_uniq_vm_name(self)
        # Create a shared config
        conf = config.VmStack(None)
        conf.store._load_from_string('''
[{vm_name}]
vm.name = {vm_name}
vm.class = ephemeral-lxd
vm.backing = byov-tests-lxd-debian
'''.format(vm_name=self.vm_name))
        conf.store.save()
        conf.store.unload()

    def test_simple_command(self):
        vm = lxd.EphemeralLxd(config.VmStack(self.vm_name))
        self.addCleanup(vm.stop)
        vm.start()
        ret, out, err = vm.shell_captured('whoami')
        self.assertEqual(0, ret)
        self.assertEqual(vm.conf.get('vm.user') + '\n', out)

# test_ssh_no_args (hard, interactive session)
# test_ssh_cant_connect (wrong host, unknown host, missing key, wrong user)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2015, 2016, 2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import errno
import io
import os
import unittest

from byov import (
    commands,
    config,
    errors,
    subprocesses,
)
from byov.tests import (
    features,
    fixtures,
)
from byov.vms import lxd


def lxd_config_get(vm_name, option):
    config_command = ['lxc', 'config', 'get', vm_name, option]
    return subprocesses.run(config_command)


class TestInfo(unittest.TestCase):

    def test_unknown(self):
        info = lxd.lxd_info('idontexist', '')
        self.assertEqual('UNKNOWN', info['state'])

    def test_running(self):
        info = lxd.lxd_info('foo', '''
Name: foo
Architecture: x86_64
Created: 2016/03/23 09:35 UTC
Status: Running
Type: persistent
Profiles: default
Pid: 7821
Processes: 20
Ips:
  eth0:	inet	192.168.0.177	vethMBG5X5
  eth0:	inet6	fe80::216:3eff:fe1e:be0f	vethMBG5X5
  lo:	inet	127.0.0.1
  lo:	inet6	::1
  lxcbr0:	inet	10.0.3.1
  lxcbr0:	inet6	fe80::1833:eff:fe84:a812
Resources:
  Processes: 18
  Disk usage:
    root: 11.24MB
  Memory usage:
    Memory (current): 77.34MB
    Memory (peak): 200.96MB
  Network usage:
    eth0:
      Bytes received: 130.75kB
      Bytes sent: 1.30kB
      Packets received: 885
      Packets sent: 885
    lo:
      Bytes received: 1.22kB
      Bytes sent: 1.22kB
      Packets received: 16
      Packets sent: 16
    lxcbr0:
      Bytes received: 0 bytes
      Bytes sent: 570 bytes
      Packets received: 0
      Packets sent: 0
''')
        self.assertEqual('RUNNING', info['state'])
        self.assertEqual('7821', info['pid'])
        self.assertEqual('192.168.0.177', info['ip'])

    def test_stopped(self):
        info = lxd.lxd_info('foo', '''
Name: foo
Status: Stopped
''')
        self.assertEqual({'state': 'STOPPED'}, info)

    def test_no_ips(self):
        # Before ips are set
        info = lxd.lxd_info('foo', '''
Name: foo
Status: Running
Ips:
(none)
''')
        self.assertEqual({'state': 'RUNNING'}, info)
        self.assertFalse('ip' in info)


@features.requires(features.lxd_client_feature)
class TestConfig(unittest.TestCase):

    def setUp(self):
        super(TestConfig, self).setUp()
        fixtures.setup_tests_config(self)
        fixtures.set_uniq_vm_name(self)
        # Create a shared config
        conf = config.VmStack(None)
        conf.store._load_from_string('''
[{vm_name}]
vm.name = {vm_name}
vm.class = lxd
'''.format(vm_name=self.vm_name))
        conf.store.save()
        conf.store.unload()

    def test_autostart(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        # A real test requires rebooting the host, we just check the default
        # value is the same as lxd itself.
        self.assertEqual(False, vm.conf.get('lxd.config.boot.autostart'))


@features.requires(features.lxd_client_feature)
class TestInit(unittest.TestCase):

    def setUp(self):
        super(TestInit, self).setUp()
        fixtures.setup_tests_config(self)
        fixtures.set_uniq_vm_name(self)
        # Create a shared config
        conf = config.VmStack(None)
        conf.store._load_from_string('''
[{vm_name}]
vm.name = {vm_name}
vm.class = lxd
'''.format(vm_name=self.vm_name))
        conf.store.save()
        conf.store.unload()

    def get_image_id(self, series, arch):
        return lxd.byov_image_name('cloudimg', series, arch)

    def test_unknown_image(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        vm.conf.set('vm.release', 'Idontexist')
        vm.conf.set('vm.architecture', 'amd64')
        with self.assertRaises(errors.CommandError) as cm:
            vm.init()
        # The error message has changed:
        # 2.0: The requested image couldn't be found.
        # 2.5: Couldn't find the requested image
        self.assertIn('he requested image', cm.exception.err)

    def test_known_image(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        vm.conf.set('vm.release', 'bionic')
        vm.conf.set('vm.architecture', 'amd64')
        self.addCleanup(vm.teardown)
        ret, out, err = vm.init()
        self.assertEqual(0, ret)
        self.assertTrue(out.startswith(('Creating {}\n'.format(
            vm.conf.get('vm.name')))))
        self.assertEqual('', err)

    def test_bad_nesting_config(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        # Short of installing testing /etc/sub[ug]id files, using an insane
        # level of nesting.
        vm.conf.set('lxd.nesting', '1024')
        vm.conf.set('vm.release', 'bionic')
        vm.conf.set('vm.architecture', 'amd64')
        with self.assertRaises(errors.ByovError) as cm:
            vm.init()
        self.assertEqual('Lxd needs more ids for 1024 levels of nesting',
                         str(cm.exception))


# FIXME: parametrized by distribution ? -- vila 2017-12-06
@features.requires(features.lxd_client_feature)
class TestSetup(unittest.TestCase):

    def setUp(self):
        super(TestSetup, self).setUp()
        fixtures.setup_tests_config(self)
        fixtures.set_uniq_vm_name(self)
        # Create a shared config
        conf = config.VmStack(self.vm_name)
        conf.set('vm.name', self.vm_name)
        conf.set('vm.class', 'lxd')
        conf.set('vm.release', 'focal')
        conf.set('vm.architecture', 'amd64')
        conf.store.save()
        conf.store.unload()
        fixtures.override_logging(self)

    def get_image_id(self, series, arch):
        return lxd.byov_image_name('cloudimg', series, arch)

    def test_usable(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        self.addCleanup(vm.teardown)
        self.addCleanup(vm.stop)
        vm.setup()
        # We should be able to ssh with the right user
        ret, out, err = vm.shell_captured('whoami')
        self.assertEqual(0, ret)
        self.assertEqual(vm.conf.get('vm.user') + '\n', out)
        # By default, nesting is not set
        ret, out, err = lxd_config_get(self.vm_name, 'security.nesting')
        self.assertEqual(0, ret)
        self.assertEqual('\n', out)
        self.assertEqual('', err)
        # Nor is lxd.remote
        self.assertIs(None, vm.econf.get('lxd.remote'))

    def test_normal_user(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        vm.conf.set('vm.user', 'tester')
        self.addCleanup(vm.teardown)
        self.addCleanup(vm.stop)
        vm.setup()
        # We should be able to ssh with the right user
        ret, out, err = vm.shell_captured('whoami')
        self.assertEqual(0, ret)
        self.assertEqual(vm.conf.get('vm.user') + '\n', out)
        # And sudo
        ret, out, err = vm.shell_captured('sudo whoami')
        self.assertEqual(0, ret)
        self.assertEqual('root\n', out)

    def test_system_user(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        vm.conf.set('vm.user', 'systester')
        vm.conf.set('vm.user.home', '/systester')
        self.addCleanup(vm.teardown)
        self.addCleanup(vm.stop)
        vm.setup()
        # We should be able to ssh with the right user
        ret, out, err = vm.shell_captured('whoami')
        self.assertEqual(0, ret)
        self.assertEqual(vm.conf.get('vm.user') + '\n', out)
        # In the right place
        ret, out, err = vm.shell_captured('pwd')
        self.assertEqual(0, ret)
        self.assertEqual('/systester\n', out)

    @features.requires(features.lxd_nesting_1)
    def test_nesting(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        vm.conf.set('lxd.nesting', '1')
        self.addCleanup(vm.teardown)
        self.addCleanup(vm.stop)
        vm.setup()
        ret, out, err = lxd_config_get(self.vm_name, 'security.nesting')
        self.assertEqual(0, ret)
        self.assertEqual('True\n', out)
        self.assertEqual('', err)

    def test_absolute_user_mounts(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        os.mkdir('testing')
        vm.conf.set('lxd.user_mounts',
                    self.uniq_dir + '/testing:/home/{vm.user}/testing')
        self.addCleanup(vm.teardown)
        self.addCleanup(vm.stop)
        vm.setup()
        ret, out, err = vm.shell_captured('touch testing/inside')
        self.assertEqual(0, ret)
        self.assertEqual('', out)
        self.assertEqual('', err)
        # The file is seen from the host
        self.assertTrue(os.path.exists('testing/inside'))

    def test_relative_user_mounts(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        os.mkdir('testing')
        vm.conf.set('lxd.user_mounts', 'testing:/home/{vm.user}/testing')
        self.addCleanup(vm.teardown)
        self.addCleanup(vm.stop)
        vm.setup()
        ret, out, err = vm.shell_captured('touch testing/inside')
        self.assertEqual(0, ret)
        self.assertEqual('', out)
        self.assertEqual('', err)
        # The file is seen from the host
        self.assertTrue(os.path.exists('testing/inside'))

    def test_homedir_user_mounts(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        os.mkdir('home/testing')
        vm.conf.set('lxd.user_mounts', '~/testing:/home/{vm.user}/testing')
        self.addCleanup(vm.teardown)
        self.addCleanup(vm.stop)
        vm.setup()
        ret, out, err = vm.shell_captured('touch testing/inside')
        self.assertEqual(0, ret)
        self.assertEqual('', out)
        self.assertEqual('', err)
        # The file is seen from the host
        self.assertTrue(os.path.exists('home/testing/inside'))

    # MISSINGTEST: What happens if two vms use the same hwaddr at the same time
    # (or not) ? -- vila 2018-04-23
    def test_hwaddr(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        vm.setup()
        # FIXME: Can we do better than eth0 ? -- vila 2018-04-24
        ret, out, err = lxd_config_get(self.vm_name, 'volatile.eth0.hwaddr')
        self.assertEqual(0, ret)
        self.assertEqual('', err)
        hwaddr = out.strip()
        ip = vm.conf.get('vm.ip')
        vm.teardown(True)
        # Arguably this can fail if the mac address is assigned by lxd to a
        # different container.
        vm.conf.set('lxd.mac.address', hwaddr)
        self.addCleanup(vm.teardown, True)
        vm.setup()
        # FIXME: The test has failed with: AssertionError: '10.242.247.98' !=
        # '10.242.247.99'. I.e. at least thrice, lxd failed to delivered the
        # same IP twice (root cause not understood yet) -- vila 2019-11-15
        self.assertEqual(ip, vm.conf.get('vm.ip'))


@features.requires(features.lxd_client_feature)
class TestCheckSubidsForMountsErrors(unittest.TestCase):
    """Test errors related to sub ids.

    Strictly speaking this should happen in a nested vm so we can change
    /etc/subuid and /ets/subgid. Doing it in the temp directory is easier.
    """
    def setUp(self):
        super(TestCheckSubidsForMountsErrors, self).setUp()
        fixtures.setup_tests_config(self)
        fixtures.set_uniq_vm_name(self)
        # Create a shared config
        conf = config.VmStack(None)
        conf.store._load_from_string('''
[{vm_name}]
vm.name = {vm_name}
vm.class = lxd
vm.release = bionic
vm.architecture = amd64
lxd.user_mounts = .:/work
'''.format(vm_name=self.vm_name))
        conf.store.save()
        conf.store.unload()

    def create_ids(self, uids, gids):
        with open(os.path.join(self.etc_dir, 'subuid'), 'w') as f:
            f.write(uids)
        with open(os.path.join(self.etc_dir, 'subgid'), 'w') as f:
            f.write(gids)

    def test_root_cant_map_uid(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        self.create_ids('', 'root:42:1')
        with self.assertRaises(errors.SubIdError) as cm:
            vm.check_subids_for_mounts(888, 0, self.etc_dir)
        self.assertEqual('uid', cm.exception.typ)
        self.assertEqual(888, cm.exception.subid)

    def test_root_cant_map_gid(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        self.create_ids('root:888:1', 'root:42:1')
        with self.assertRaises(errors.SubIdError) as cm:
            vm.check_subids_for_mounts(888, 888, self.etc_dir)
        self.assertEqual('gid', cm.exception.typ)
        self.assertEqual(888, cm.exception.subid)

    def test_root_can_map(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        self.create_ids('root:888:1', 'root:999:1')
        vm.check_subids_for_mounts(888, 999, self.etc_dir)


# FIXME: parametrized by distribution -- vila 2017-12-06
class TestEphemeralLXD(unittest.TestCase):

    def setUp(self):
        super(TestEphemeralLXD, self).setUp()
        fixtures.setup_tests_config(self)
        features.requires_existing_vm(self, 'byov-tests-lxd-debian')
        fixtures.set_uniq_vm_name(self)
        # Create a shared config
        conf = config.VmStack(None)
        conf.store._load_from_string('''
[{vm_name}]
vm.name = {vm_name}
vm.class = ephemeral-lxd
vm.backing = byov-tests-lxd-debian
'''.format(vm_name=self.vm_name))
        conf.store.save()
        conf.store.unload()

    def test_start(self):
        vm = lxd.EphemeralLxd(config.VmStack(self.vm_name))
        self.addCleanup(vm.stop)
        vm.start()
        self.assertEqual('RUNNING', vm.state())

    def test_stop(self):
        vm = lxd.EphemeralLxd(config.VmStack(self.vm_name))

        def cleanup():
            # vm.start() may fail if no IP can't be acquired (for example). In
            # that case, the container runs and needs to be stopped. If the
            # test succeeds, the vm has already been stopped. This awkward
            # cleanup avoid leaks.
            if vm.state() == 'RUNNING':
                vm.stop()

        self.addCleanup(cleanup)
        vm.start()
        vm.stop()
        # There is a tiny window during which the container can be seen STOPPED
        # before being deleted
        self.assertIn(vm.state(), ('UNKNOWN', 'STOPPED'))

    def test_forbid_lxd_user_mounts_in_both(self):
        conf = config.VmStack('byov-tests-lxd-debian')
        conf.set('lxd.user_mounts', 'home:/work')
        conf.store.save()
        vm = lxd.EphemeralLxd(config.VmStack(self.vm_name))
        vm.conf.set('lxd.user_mounts', '.:/ephemeral')
        # If the backing container has specified lxd.user_mounts, the ephemeral
        # cannot.
        with self.assertRaises(errors.ByovError) as cm:
            vm.start()
        self.assertEqual(
            'Backing vm byov-tests-lxd-debian already has mounts',
            str(cm.exception))

    def test_ephemeral_mounts(self):
        vm = lxd.EphemeralLxd(config.VmStack(self.vm_name))
        os.mkdir('testing')
        vm.conf.set('lxd.user_mounts',
                    self.uniq_dir + '/testing:/home/{vm.user}/testing')
        self.addCleanup(vm.stop)
        vm.start()
        ret, out, err = vm.shell_captured('touch testing/inside')
        self.assertEqual(0, ret)
        self.assertEqual('', out)
        self.assertEqual('', err)
        # The file is seen from the host
        self.assertTrue(os.path.exists('testing/inside'))


class TestNextSubuids(unittest.TestCase):

    def setUp(self):
        super(TestNextSubuids, self).setUp()
        fixtures.isolate_from_disk(self)

    def assertIdRanges(self, expected, content):
        with open(os.path.join(self.etc_dir, 'subuid'), 'w') as f:
            f.write(content)
        self.assertEqual(expected, list(lxd.next_subids('uid', self.etc_dir)))

    def test_unknown(self):
        with self.assertRaises(IOError) as cm:
            list(lxd.next_subids('uid', self.etc_dir))
        # aka FileNotFound (python2 specific)
        self.assertEqual(2, cm.exception.args[0])
        self.assertEqual(errno.ENOENT, cm.exception.errno)

    def test_empty(self):
        self.assertIdRanges([], '')
        self.assertIdRanges([], '\n\n')

    def test_garbage(self):
        self.assertIdRanges([], 'root:foo:2')
        self.assertIdRanges([], 'root:2:foo')

    def test_no_newline(self):
        self.assertIdRanges([('root', 1000, 1)], 'root:1000:1')

    def test_ignore_comments(self):
        self.assertIdRanges([('root', 1000, 1), ('foo', 12, 42)],
                            'root:1000:1\n# foo\nfoo:12:42')


class TestCheckIdFor(unittest.TestCase):

    def setUp(self):
        super(TestCheckIdFor, self).setUp()
        fixtures.isolate_from_disk(self)

    def create_ids(self, typ, ids):
        with open(os.path.join(self.etc_dir, 'sub{}'.format(typ)), 'w') as f:
            f.write(ids)

    def test_present(self):
        self.create_ids('uid', 'root:1000:1')
        self.assertTrue(lxd.check_id_for('root', 1000, 'uid', self.etc_dir))

    def test_absent_from_subgid(self):
        self.create_ids('gid', '')
        self.assertFalse(lxd.check_id_for('root', 1000, 'gid', self.etc_dir))

    def test_absent_from_subuid(self):
        self.create_ids('uid', '')
        self.assertFalse(lxd.check_id_for('root', 1000, 'uid', self.etc_dir))


class TestCheckNesting(unittest.TestCase):

    def setUp(self):
        super(TestCheckNesting, self).setUp()
        fixtures.isolate_from_disk(self)

    def create_ids(self, uids, gids):
        with open(os.path.join(self.etc_dir, 'subuid'), 'w') as f:
            f.write(uids)
        with open(os.path.join(self.etc_dir, 'subgid'), 'w') as f:
            f.write(gids)

    def test_no_nesting(self):
        self.create_ids('root:100000:65536\nlxd:100000:65536\n',
                        'root:100000:65536\nlxd:100000:65536\n')
        self.assertTrue(lxd.check_nesting(0, self.etc_dir))

    def test_nesting_one(self):
        self.create_ids('root:100000:131072\nlxd:100000:131072\n',
                        'root:100000:131072\nlxd:100000:131072\n')
        self.assertTrue(lxd.check_nesting(1, self.etc_dir))

    def test_nesting_two(self):
        self.create_ids('root:100000:196608\nlxd:100000:196608\n',
                        'root:100000:196608\nlxd:100000:196608\n')
        self.assertTrue(lxd.check_nesting(2, self.etc_dir))

    def test_not_enough(self):
        self.create_ids('root:100000:131072\nlxd:100000:131072\n',
                        'root:100000:131072\nlxd:100000:131072\n')
        self.assertFalse(lxd.check_nesting(2, self.etc_dir))


class TestRemoteAdd(unittest.TestCase):

    def setUp(self):
        super(TestRemoteAdd, self).setUp()
        fixtures.setup_tests_config(self)
        features.requires_existing_vm(self, 'byov-lxd-client')
        features.requires_existing_vm(self, 'byov-lxd-server')
        fixtures.set_uniq_vm_name(self)
        self.client_name = 'client-' + self.vm_name
        self.server_name = 'server-' + self.vm_name
        # Create a shared config
        conf = config.VmStack(None)
        conf.store._load_from_string('''
[{client}]
vm.name = {client}
vm.class = ephemeral-lxd
vm.backing = byov-lxd-server
[{server}]
vm.name = {server}
vm.class = ephemeral-lxd
vm.backing = byov-lxd-server
'''.format(client=self.client_name, server=self.server_name))
        conf.store.save()
        conf.store.unload()
        self.client = self.start_ephemeral(self.client_name)
        self.server = self.start_ephemeral(self.server_name)
        self.out = io.StringIO()
        self.err = io.StringIO()
        fixtures.override_logging(self)

    def start_ephemeral(self, vm_name):
        vm = lxd.EphemeralLxd(config.VmStack(vm_name))
        self.addCleanup(vm.stop)
        vm.start()
        return vm

    def shell_on_vm(self, vm_name, args, out=None, err=None):
        if out is None:
            out = self.out
        if err is None:
            err = self.err
        full_cmd = ['shell', vm_name] + args
        ret = commands.run(full_cmd, out=out, err=err)
        msg = 'Cmd {} failed with {}:\n\tstdout: {}\n\tstderr: {}'
        self.assertEqual(0, ret, msg.format(
            full_cmd, ret, out.getvalue(),
            err.getvalue()))
        return ret, out.getvalue(), err.getvalue()

    def shell_on_server(self, args, out=None, err=None):
        return self.shell_on_vm(self.server_name, args, out=out, err=err)

    def test_add_remote(self):
        self.shell_on_server(['lxc config set core.https_address :8443'])
        self.shell_on_server(['lxc config set core.trust_password s3cr3t'])
        server_ip = self.server.conf.get('vm.ip')
        cmd_args = ['remote', 'add', self.server_name,
                    '{}:8443'.format(server_ip),
                    '--password', 's3cr3t', '--accept-certificate']
        ret, out, err = self.client.shell_captured('lxc', *cmd_args)
        msg = 'remote add {} failed with {}:\n\tstdout: {}\n\tstderr: {}'
        self.assertEqual(0, ret, msg.format(cmd_args, ret, out, err))
        _, out, _ = self.client.shell_captured('lxc', 'remote', 'list')
        stem = self.server_name
        self.assertTrue(stem in out, 'Cannot find {} in {}'.format(stem, out))
        stem = 'https://' + server_ip + ':8443'
        self.assertTrue(stem in out, 'Cannot find {} in {}'.format(stem, out))


class TestRemote(unittest.TestCase):

    def setUp(self):
        super(TestRemote, self).setUp()
        fixtures.setup_tests_config(self)
        fixtures.set_uniq_vm_name(self)
        self.lxd_remote = features.requires_remote_lxd(self)
        # Create a shared config
        conf = config.VmStack(self.vm_name)
        conf.set('vm.name', self.vm_name)
        conf.set('vm.class', 'lxd')
        conf.set('vm.distribution', 'ubuntu')
        conf.set('vm.release', 'bionic')
        conf.set('vm.architecture', 'amd64')
        conf.set('lxd.remote', self.lxd_remote)
        conf.store.save()
        conf.store.unload()
        fixtures.override_logging(self)

    def test_setup(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        self.addCleanup(vm.teardown)
        self.addCleanup(vm.stop)
        vm.setup()
        # We should be able to ssh with the right user
        ret, out, err = vm.shell_captured('whoami')
        self.assertEqual(0, ret)
        self.assertEqual(vm.conf.get('vm.user') + '\n', out)
        # The lxd.remote option should be saved.
        self.assertEqual(self.lxd_remote, vm.econf.get('lxd.remote'))

    def test_mount_map_defaults(self):
        # lxd.user_mounts are a bit tricky: the host uid/gid is the remote one
        # so it needs to be overridable.
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        mount_map = vm.get_mount_id_map()
        # The defaults are a bit ubuntu/debian specific (but they may also be
        # valid for other distros (we're not there yet)).
        self.assertEqual(
            ('uid 1000-1000 1000-1000', 'gid 1000-1000 1000-1000'),
            mount_map)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest


from byov import logs


class TestIpAddrRegexp(unittest.TestCase):

    def setUp(self):
        super(TestIpAddrRegexp, self).setUp()
        self.re = logs.ip_address_re

    def assertMatches(self, ip_addr):
        match = self.re.match(ip_addr)
        self.assertIsNotNone(match)

    def assertDontMatch(self, ip_addr):
        match = self.re.match(ip_addr)
        self.assertIsNone(match)

    def test_matches(self):
        self.assertMatches('192.168.0.12')
        self.assertMatches('19.16.0.0')
        self.assertMatches('1.1.1.1')

    def test_dont_match(self):
        self.assertDontMatch('')
        self.assertDontMatch('1924.168.0.0')
        self.assertDontMatch('192.1682.0.0')
        self.assertDontMatch('192.168..0')


class TestMacAddrRegexp(unittest.TestCase):

    def setUp(self):
        super(TestMacAddrRegexp, self).setUp()
        self.re = logs.mac_address_re

    def assertMatches(self, mac_addr):
        match = self.re.match(mac_addr)
        self.assertIsNotNone(match)

    def assertDontMatch(self, mac_addr):
        match = self.re.match(mac_addr)
        self.assertIsNone(match)

    def test_matches(self):
        self.assertMatches('00:16:3e:38:8d:7e')
        self.assertMatches('00:16:3e:87:7b:38')
        self.assertMatches('00:50:e4:70:b3:23')
        self.assertMatches('00:24:b2:38:f6:d6')
        self.assertMatches('00:16:CB:9E:F0:A1')
        self.assertMatches('08:00:27:9B:A0:2C')
        self.assertMatches('10:9A:DD:A7:08:DA')
        self.assertMatches('00:1D:0F:B0:EB:3A')
        self.assertMatches('52:54:00:04:7B:45')

    def test_dont_match(self):
        self.assertDontMatch('')
        self.assertDontMatch('08:00:27:2d:64:')
        self.assertDontMatch(':00:27:52:27:b8')
        self.assertDontMatch('c4:3d:c7:55:5c:8')
        self.assertDontMatch('0:16:3e:c5:62:08')


class TestIfaceRegexp(unittest.TestCase):

    def setUp(self):
        super(TestIfaceRegexp, self).setUp()
        self.re = logs.InterfaceRegexp('(eth0|ens3)')

    def assertMatches(self, expected, line):
        match = self.re.match(line)
        self.assertIsNotNone(match)
        iface, ip, mask, mac = match.groups()
        self.assertEqual(expected, (iface, ip, mask, mac))

    def test_matches_quantal(self):
        line = 'ci-info: |  {}  | True | {} | {} | {} |'
        addrs = ('eth0', '192.168.0.119', '255.255.255.0', '00:16:3e:38:8d:7e')
        self.assertMatches(addrs, line.format(*addrs))

    def test_matches_xenial(self):
        line = 'ci-info: |  {}  | True | {} | {} | {} |'
        addrs = ('ens3', '192.168.0.119', '255.255.255.0', '00:16:3e:38:8d:7e')
        self.assertMatches(addrs, line.format(*addrs))

    def test_matches_precise(self):
        line = 'ci-info: {}  : 1  {}   {}   {}'
        addrs = ('eth0', '192.168.0.119', '255.255.255.0', '00:16:3e:38:8d:7e')
        self.assertMatches(addrs, line.format(*addrs))



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals
import errno
import getpass
import logging
import os
import sys

from byoc import (
    options,
    registries,
)
from byov import (
    subprocesses,
    timeouts,
)

# Official API
Option = options.Option
ListOption = options.ListOption
PathOption = options.PathOption
MANDATORY = options.MANDATORY
OptionMandatoryValueError = options.errors.OptionMandatoryValueError
option_registry = options.option_registry
register = options.register


logger = logging.getLogger(__name__)


class PackageListOption(options.ListOption):
    """A package list option definition.

    This possibly expands '@filename' replacing it with the file content.
    """

    def __init__(self, *args, **kwargs):
        # Foce invalid values to be an error (and forbids overriding it) to
        # catch invalid file names
        super().__init__(*args, invalid='error', **kwargs)

    def from_unicode(self, string):
        values = super().from_unicode(string)
        if not values:
            return values
        converted = []
        for v in values:
            if v.startswith('@'):
                packages_path = os.path.expanduser(v[1:])
                try:
                    # FIXME: A bit more care should be taken about interpreting
                    # relative paths: are they relative to the config file
                    # (they probably should) ?  Also, the current
                    # implementation swallow the ValueError raised below,
                    # turning it into an OptionValueError instead, losing the
                    # precision if multiple files are involved
                    # -- vila 2015-06-18
                    with open(packages_path) as f:
                        for package in f.read().splitlines():
                            comment = package.find('#')
                            if comment >= 0:
                                # Delete comments
                                package = package[0:comment]
                                package = package.rstrip()
                            if package == '':
                                # Ignore empty lines
                                continue
                            converted.append(package)
                except IOError as e:
                    # python2 does not provide FileNotFoundError
                    if e.errno == errno.ENOENT:
                        raise ValueError('{} does not exist'.format(v[1:]))
                    else:
                        raise
            else:
                converted.append(v)
        return converted


class LxdUserMountsOption(options.ListOption):
    """A list option definition for lxd user mounts.

    The list should contain pairs of host/vm paths.
    The host paths will be mounted at the vm paths with the user uid/gid.
    """

    def __init__(self, *args, **kwargs):
        # Foce invalid values to be an error (and forbids overriding it) to
        # catch invalid file names
        super().__init__(
            *args, invalid='error', **kwargs)

    def from_unicode(self, string):
        values = super().from_unicode(string)
        if not values:
            return values
        converted = []
        for pair in values:
            try:
                host, vm = pair.split(':')
            except ValueError:
                raise ValueError('Invalid path pair in: {}'.format(pair))
            host = os.path.expanduser(host)
            if not os.path.isabs(host):
                # We're on thin ice if byovm is not run from the directory
                # where the byov.conf file is...
                host = os.path.realpath(os.path.join(os.getcwd(), host))
            if not os.path.isabs(vm):
                raise ValueError('{} must be absolute'.format(vm))
            converted.append((host, vm))
        return converted


class DockerMountsOption(options.ListOption):
    """A list option definition for docker mounts.

    This relies on VOLUMEs being defined in the docker image.
    The list should contain type-specific tuples.
    """
    def __init__(self, *args, **kwargs):
        # Foce invalid values to be an error (and forbids overriding it) to
        # catch invalid directories
        super().__init__(
            *args, invalid='error', **kwargs)

    def parse_bind(self, mnt):
        tmpl = '--mount=type=bind,src={src},dst={dst}'
        valid_shares = ('rprivate', 'private', 'rshared', 'shared', 'rslave',
                        'slave')
        values = mnt.split(':')
        src, dst = values[1:3]
        src = os.path.expanduser(src)
        if not os.path.isabs(src):
            # We're on thin ice if byovm is not run from the directory
            # where the byov.conf file is...
            src = os.path.realpath(os.path.join(os.getcwd(), src))
        if not os.path.isabs(src):
            raise ValueError('{} must be absolute'.format(src))
        ro = False
        share = None
        if len(values) > 3:
            ro = values[3]
            if ro != 'readonly':
                raise ValueError('Only the readonly value is supported')
            if len(values) > 4:
                share = values[4]
                if share not in valid_shares:
                    raise ValueError(share + 'is not in ' + valid_shares)
        res = tmpl.format(**locals())
        if ro:
            res += ',readonly'
        if share:
            res += ',bind-propagation=' + share
        return res

    def parse_tmpfs(self, mnt):
        tmpl = '--mount=type=tmpfs,dst={dst}'
        values = mnt.split(':')
        dst = values[1]
        if len(values) > 2:
            size = int(values[2])
            if len(values) > 3:
                mode = values[3]
        res = tmpl.format(**locals())
        if size:
            res += ',tmpfs-size={}'.format(size)
        if mode:
            res += ',tmpfs-mode=' + mode
        return res

    def parse_volume(self, mnt):
        tmpl = '--mount=type=volume,src={src},dst={dst}'
        values = mnt.split(':')
        src, dst = values[1:3]
        ro = False
        if len(values) > 3:
            ro = values[3]
            if ro != 'readonly':
                raise ValueError('Only the readonly value is supported')
        res = tmpl.format(**locals())
        if len(values) > 3:
            ro = values[3]
            if ro != 'readonly':
                raise ValueError('Only the readonly value is supported')
        if ro:
            res += ',readonly'
        return res

    def from_unicode(self, string):
        values = super().from_unicode(string)
        if not values:
            return values
        converted = []
        for mnt in values:
            typ = mnt.split(':', 1)[0]
            if typ not in ('bind, tmpfs, volume'):
                raise ValueError('Unknown type in ' + mnt)
            if typ == 'bind':
                converted.append(self.parse_bind(mnt))
            elif typ == 'volume':
                converted.append(self.parse_volume(mnt))
            elif typ == 'tmpfs':
                converted.append(self.parse_tmpfs(mnt))
        return converted


class TimeoutsOption(options.ListOption):
    """Timeouts definition for an exponential backoff on retries."""

    def __init__(self, name, help_string, **kwargs):
        # Add help boilerplate
        help_string += '''

The (first, up_to, retries) tuple is defined as:
- first: seconds to wait after the first attempt
- up_to: seconds after which to give up
- retries: how many attempts after the first try
'''
        # Force invalid values to be an error (and forbids overriding it) to
        # catch invalid values
        super().__init__(name, help_string, invalid='error', **kwargs)

    def from_unicode(self, string):
        values = super().from_unicode(string)
        if not values:
            return values
        if len(values) != 3:
            raise ValueError('{} has invalid length: {}'.format(
                values, len(values)))
        first, up_to, retries = values
        first = float(first)
        up_to = float(up_to)
        retries = int(retries)
        return timeouts.ExponentialBackoff(first, up_to, retries)


register(options.Option('vm', default=None,
                        help_string='''\
The name space defining a virtual machine.

This option is a place holder to document the options that defines a virtual
machine and the options defining the infrastructure used to manage them all.

For libvirt based vms, the definition of a vm is stored in an xml file under
'/etc/libvirt/qemu/{vm.name}.xml'. This is under the libvirt package control
and is out of scope for byov.

There are 3 other significant files used for a given libvirt vm:

- a disk image mounted at '/' from '/dev/sda1':
  '{libvirt.images_dir}/{vm.name}.qcow2'

- an iso image available from '/dev/sdb' labeled 'cidata':
  {libvirt.images_dir}/{vm.name}.seed which contains the cloud-init data used
  to configure/install/update the vm.

- a console: {libvirt.images_dir}/{vm.name}.console which can be 'tail -f'ed
  from the host.

The data used to create the seed above are stored in a vm specific
configuration directory for easier debug and reference:
- {vm.config_dir}/user-data
- {vm.config_dir}/meta-data
- {vm.config_dir}/ecdsa
- {vm.config_dir}/ecdsa.pub
'''))

# The directory where we store vm files related to their configuration with
# cloud-init (user-data, meta-data, ssh server keys).
register(options.PathOption('vm.vms_dir', default='~/.config/byov',
                            help_string='''\
Where vm related config files are stored.

This includes user-data and meta-data for cloud-init and ssh server keys.

This directory must exist.

Each vm get a specific directory (automatically created) there based on its
name.
'''))
# The VM classes are registered where/when needed
vm_class_registry = registries.Registry()


register(options.RegistryOption('vm.class', registry=vm_class_registry,
                                default=options.MANDATORY,
                                help_string='''\
The virtual machine technology to use.'''))
# The ubiquitous vm name
register(options.Option('vm.name', default=None, invalid='error',
                        help_string='''\
The vm name, used as a prefix for related files.'''))
# The next most important bits: distribution and release
register(options.Option('vm.distribution', default='ubuntu', invalid='error',
                        help_string='''The distribution name.'''))
register(options.Option('vm.release', default=None, invalid='error',
                        help_string='''The release name.'''))
register(options.Option('vm.architecture', default=None, invalid='error',
                        help_string='''The vm architecture (cpu model).'''))
register(options.Option('vm.published_as',
                        default=options.MANDATORY,
                        help_string='''\
The name used to publish the stopped vm as an image.'''))
register(options.Option('vm.fqdn', default=None,
                        help_string='''\
The fully qualified domain name for the vm.'''))
register(options.Option('vm.manage_etc_hosts', default='True',
                        help_string='''\
Whether cloud-init should manage /etc/hosts or not.

This is rarely needed but can be set to 'localhost' to avoid conflicts with
other packages willing to modify /etc/hosts (puppet is a known case).'''))

# The third important piece to define a vm: where to store files like the
# console, the user-data and meta-data files, the ssh server keys, etc.
register(options.PathOption('vm.config_dir', default='{vm.vms_dir}/{vm.name}',
                            invalid='error',
                            help_string='''\
The directory where files specific to a vm are stored.

This includes the user-data and meta-data files used at install time (for
reference and easier debug) as well as the optional ssh server keys.

By default this is {vm.vms_dir}/{vm.name}. You can put it somewhere else by
redefining it as long as it ends up being unique for the vm.

{vm.vms_dir}/{vm.release}/{vm.name} may better suit your taste for example.
'''))
# The options defining the vm physical characteristics
register(options.Option('vm.cpus', default='1', help_string='''\
The number of cpus.'''))
register(options.Option('vm.ram_size', default='2048',
                        help_string='''The ram size in megabytes.'''))
register(options.Option('vm.disk_size', default='8G',
                        help_string='''The disk image size in bytes.

Optional suffixes "k" or "K" (kilobyte, 1024) "M" (megabyte, 1024k) "G"
(gigabyte, 1024M) and T (terabyte, 1024G) are supported.
'''))
register(options.Option('vm.meta_data', default='''\
instance-id: {vm.name}
local-hostname: {vm.name}
''',
                        invalid='error',
                        help_string='''\
The meta data for cloud-init to put in the seed.'''))

register(options.Option('vm.user',
                        default='{{vm.distribution}.user}',
                        help_string='''The user in the vm.'''))
register(options.Option('vm.user.home', default='/home/{vm.user}',
                        help_string='''\
Home directory for the user.

Rarely needed but when it cannot be guessed, a life saver.
'''))
register(options.Option('vm.user.shell',
                        default='{{vm.distribution}.user.shell}',
                        help_string='''\
The user's login shell.

The default is system/distribution/image specific.
'''))
register(options.Option('vm.user.system', default=False, invalid='error',
                        from_unicode=options.bool_from_store,
                        help_string='''\
Is {vm.user} a system user.
'''))
register(options.ListOption('vm.user.sudo',
                            default='ALL=(ALL) NOPASSWD:ALL',
                            help_string='''\
A list of sudo strings to be installed for {vm.user}.
'''))
# Some bits that may be added to user-data but are optional

register(options.ListOption('vm.chpasswd',
                            default=None,
                            help_string='''\
A list of tuples (user:pass) defining passwords.

The passwords can be hashed with mkpasswd but as the cloud-init explains: "If
you do not fully trust the medium over which your cloud-config will be
transmitted, then you should use SSH authentication only.
'''))
register(options.Option('vm.locale', default='en_US.UTF-8',
                        help_string='''The locale to use.'''))
# FIXME: This could default to the package manager associated option
# -- vila 2021-12-30
register(PackageListOption('vm.packages', default='',
                           help_string='''\
A list of package names to be installed.

If the package name starts with '@', it is interpreted as a file name
and its content, one package by line, is inserted in the list.
'''))
register(options.Option('vm.update', default=False,
                        from_unicode=options.bool_from_store,
                        help_string='''Whether or not the vm should be updated.

Both apt-get update and apt-get dist-upgrade are called if this option is set.
'''))
register(options.Option('vm.poweroff', default=False,
                        from_unicode=options.bool_from_store,
                        help_string='''\
Whether or not the vm should stop at the end of the installation.'''))
# The options to control vm.backing reuse
register(options.Option('vm.setup.digest', default=None,
                        help_string='''\
A digest of all option values defining the vm.

This is produced by byov itself and should not be set manually.
'''))
register(options.ListOption('vm.setup.digest.options',
                            default='vm.name, vm.class,'
                            'vm.distribution, vm.release, vm.architecture,'
                            'vm.user, {vm.distribution}.user,'
                            'vm.user.system, vm.user.home, vm.user.sudo,'
                            'vm.fqdn, vm.manage_etc_hosts,'
                            'vm.cpus, vm.ram_size, vm.disk_size,'
                            'vm.meta_data,'
                            'vm.locale,'
                            'vm.packages,'
                            'pip.packages,'
                            'vm.setup.hook,'
                            'vm.start.hook,'
                            'vm.uploaded_scripts.guest_dir,'
                            'vm.final_message,'
                            'launchpad.login,'  # Ugly duck
                            'gitlab.login',  # Ugly duck
                            help_string='''\
A list of options that are used to define a vm.

The values for these options are hashed to produce 'vm.setup.digest'.
'''))
# The scripts that are executed at the end of cloud-init setup
register(options.PathOption('vm.root_script', default=None,
                            help_string='''\
The path to a script executed as root at the end of cloud-init setup.

This script is executed before {vm.user_script}.
'''))
register(options.PathOption('vm.user_script', default=None,
                            help_string='''\
The path to a script executed as the default user at the end of cloud-init
setup.

This script is excuted after {vm.root_script}.
'''))
register(options.ListOption('vm.uploaded_scripts', default=None,
                            help_string='''\
A list of scripts to be uploaded to the guest.

Scripts can use config options from their vm, they will be expanded before
upload. All scripts are uploaded into {vm.uploaded_scripts.guest_dir} under
their base name.
'''))
register(options.Option('vm.uploaded_scripts.guest_dir',
                        default='~{{vm.distribution}.user}/bin',
                        help_string='''\
Where {vm.uploaded_scripts} are uploaded on the guest.'''))
register(options.Option('vm.setup.hook', default=None,
                        help_string='''\
A command that is executed *on the host* at the end of the setup.

If options expansion is needed, '@script' will expand the options inside
'script' before execution.

Note: This runs after 'vm.packages' have been installed and before
'vm.setup_scripts'.  '''))
register(options.ListOption('vm.setup_scripts', default=None,
                            help_string='''\
A list of scripts to be executed on the guest to finalize the setup.

The scripts are executed over ssh, in the user home directory.
Scripts can use config options from their vm, they will be expanded before
upload.
'''))
register(options.Option('vm.final_message',
                        default=None,
                        help_string='''\
The cloud-init final message, a release-specific message is used if
none is provided to address compatibility with Ubuntu precise.'''))
register(options.Option('vm.start.hook', default=None,
                        help_string='''\
A command that is executed *on the host* before starting the vm.

If options expansion is needed, '@script' will expand the options inside
'script' before execution.
'''))

######################################################################
# scaleway options
######################################################################
register(options.Option('scaleway.access_key',
                        default=options.MANDATORY,
                        help_string='''The scaleway access key.

See https://cloud.scaleway.com/#/credentials for details.
'''))
register(options.Option('scaleway.token',
                        default=options.MANDATORY,
                        help_string='''The scaleway authorization token.

See https://cloud.scaleway.com/#/credentials for valid ones.
'''))
register(options.Option('scaleway.public_ip',
                        default=None,
                        help_string='''The reserved ip to use.

The IP address to assign to the vm. It should be an existing reserved IP, not
already used by an existing server.
'''))
register(options.Option('scaleway.region_name',
                        default=options.MANDATORY,
                        help_string='''The scaleway region name.
See https://scaleway.com.
'''))
register(options.Option('scaleway.flavor',
                        default=options.MANDATORY,
                        help_string='''\
A scaleway commercial type for the server.

See https://scaleway.com.
'''))
register(options.Option('scaleway.boot.local',
                        from_unicode=options.bool_from_store,
                        default=None,
                        help_string='''\
Boot the kernel on the first volume.

Only works for some types of server so opt-in only for now.
'''))
register(options.Option('scaleway.image.bootstrap',
                        from_unicode=options.bool_from_store,
                        default='false',
                        help_string='''\
Disable cloud-init check when creating scaleway images.

Images provided by scaleway don't include cloud-init. This option disables the
checks for cloud-init completion.
'''))
# FIXME: scaleway.bootscript is missing -- vila 2018-01-24
register(options.Option(
    'scaleway.image',
    default='{vm.distribution}/{vm.release}/{vm.architecture}',
    help_string='''\
The image to boot from.

See https://www.scaleway.com/imagehub/.
'''))
# It could take up to 20 mins in the worst cases (especially when creating
# several servers at once).
register(options.Option('scaleway.poweron_timeout', default='1800',
                        from_unicode=options.float_from_store,
                        help_string='''\
Max time to power on a scaleway server (in seconds).'''))
register(options.Option('scaleway.poweroff_timeout', default='600',
                        from_unicode=options.float_from_store,
                        help_string='''\
Max time to power off a scaleway server (in seconds).'''))
register(options.Option('scaleway.terminate_timeout', default='300',
                        from_unicode=options.float_from_store,
                        help_string='''\
Max time to terminate a scaleway server (in seconds).'''))
# FIXME: It takes around 3 minutes to boot and get a dynamic IP for C1, check
# whether it's faster with a static one -- vila 2018-01-14
register(TimeoutsOption('scaleway.setup_ip.timeouts',
                        default='60, 600, 20',
                        help_string='''\
When waiting for scaleway to setup an IP.'''))
register(TimeoutsOption('scaleway.setup_ssh.timeouts', default='0, 180, 20',
                        help_string='''\
When waiting for scaleway to setup ssh.'''))
register(TimeoutsOption('scaleway.cloud_init.timeouts', default='0, 240, 20',
                        help_string='''\
When waiting for cloud-init completion.
'''))
register(options.ListOption('scaleway.setup.digest.options',
                            default='scaleway.flavor, scaleway.image,'
                            'scaleway.bootscript, scaleway.isntance_id',
                            help_string='''\
A list of scaleway related options that are used to define a vm.

The values for these options are hashed to produce 'vm.setup.digest'.
'''))
register(options.Option(
    'scaleway.compute.url',
    default='https://cp-{scaleway.region_name}.scaleway.com',
    help_string='''The scaleway compute api url.

See https://developer.scaleway.com.
'''))
register(TimeoutsOption('scaleway.compute.timeouts',
                        default='2, 30, 10',
                        help_string='''\
When talking to the scaleway compute server.'''))

######################################################################
# qemu options
######################################################################

register(options.PathOption('qemu.image',
                            default='''{qemu.images.dir}/vms/{vm.name}''',
                            help_string='''\
The path where the boot image is stored.

This defines the name space for the vms disk images using the vm name.

Most image commands end up producing an image in that name space.
'''))

register(options.ListOption('qemu.image.setup',
                            default='download,convert,resize',
                            help_string='''\
A list of qemu image commands to execute before setup.

Supported commands (see corresponding help): {qemu.download}, {qemu.convert},
{qemu.clone}, {qemu.copy}, {qemu.create}, {qemu.resize}.
'''))
register(options.ListOption('qemu.image.teardown',
                            default='convert',
                            help_string='''\
A list of qemu image commands to execute after teardown. The commands accepted
are the same as for {qemu.image.setup} but they are executed in the reverse
order at teardown time.

Depending on the workflow, .setup and .teardown can be different to give
flexibility. Using 'download,clone' for setup and 'clone' for teardown will
keep the downloaded images after the first download. Using 'download,clone' in
teardown will force a re-download.
'''))

register(options.Option('qemu.download.url',
                        default='''\
http://cloud-images.ubuntu.com/{vm.release}/current/\
{vm.release}-server-cloudimg-{vm.architecture}.img''',
                                help_string='''\
The url where the image can be downloaded from.'''))
register(options.PathOption('qemu.download.dir',
                            default='~/.cache/byov/images',
                            help_string='''Where downloads end up.'''))
register(options.PathOption('qemu.download.path',
                            default='''\
{qemu.download.dir}/{vm.distribution}/{vm.release}-{vm.architecture}''',
                            help_string='''\
The path where the image is downloaded.

This defines the name space for the downloads using distribution, release and
architecture.
'''))

register(options.ListOption('qemu.clone', separator=' ',
                            default='''\
qemu-img create -f qcow2 -b {source} {target}''',
                            help_string='''\
Create target ({qemu.image}) on top of source ({vm.published_as}
in {vm.backing}).
'''))
register(options.ListOption('qemu.convert', separator=' ',
                            default='''\
qemu-img convert -O qcow2 {source} {target}''',
                            help_string='''\
Convert source ({qemu.download.path}) into target ({qemu.image}).
'''))
register(options.ListOption('qemu.copy', separator=' ',
                            default='''\
cp {source} {target}''',
                            help_string='''\
Copy source ({qemu.download.path}) into target ({qemu.image}).
'''))
register(options.ListOption('qemu.create', separator=' ',
                            default='''\
qemu-img create -f qcow2 {target}''',
                            help_string='''\
Create image at target ({qemu.image}).
'''))
register(options.ListOption('qemu.download', separator=' ',
                            default='''\
wget {source} -nv --output-document {target}''',
                            help_string='''\
Downloads source ({qemu.download.url}) into target ({qemu.download.path}).
The download does not happen if {qemu.download.path} already exists.
'''))
# FIXME: Could each command define its own env so run_cmd() can use
# that instead of the hard-coded source/target ? -- vila 2019-12-07
# Abusing {source} to give access to disk size
register(options.ListOption('qemu.resize', separator=' ',
                            default='''\
qemu-img resize {target} {source}''',
                            help_string='''\
Resize image at target ({qemu.image}) with source ({vm.disk_size}).
'''))
register(options.PathOption('qemu.images.dir',
                            default=os.path.expanduser(
                                '~/.local/share/byov'),
                            help_string='''\
Where qemu vm disk images are stored.'''))
register(options.ListOption('qemu.disks', separator=' ',
                            default='',
                            help_string='''\
Additional disks definitions.

This is added to the qemu command under user responsibility.
'''))

# https://wiki.archlinux.org/index.php/QEMU#User-mode_networking has all the
# good bits about network setup for qemu
# https://virt.kernelnewbies.org/MacVTap for macvtap
# https://seravo.fi/2012/virtualized-bridged-networking-with-macvtap
# https://turlucode.com/qemu-kvm-bridged-networking/
# https://wiki.archlinux.org/index.php/QEMU#Bridged_networking_using_qemu-bridge-helper
register(options.ListOption('qemu.networks', separator=' ',
                            default='''\
-device e1000,netdev=user.0 \
-netdev user,id=user.0,hostfwd=tcp::{qemu.ssh.localhost.port}-:22''',
                            help_string='''\
The network definitions provided to qemu.

This can be specialized for each machine but the default should work for all
setups with the caveat that only ssh access is provided by forwarding a
localhost port (qemu.ssh.localhost.port) that needs to be set.

Alternatively, a bridge (br0 below) can be created on the host. From there,
qemu.networks can be set to:

  qemu.networks = -net bridge,br=br0 -net nic,macaddr=xx:xx:xx:xx:xx:xx

'''))
register(options.Option('qemu.ssh.localhost.port', default=None,
                        help_string='''\
When set, ssh connect to localhost on that port.

This redirection should appear in qemu.networks.
'''))
register(options.Option('qemu.bridge', default=None,
                        help_string='''\
To use in qemu.networks when a bridge is available on the host as:

  -net bridge,br={qemu.bridge} -net nic,macaddr={qemu.mac.address}

'''))


def mac_from_pid(conf=None):
    """Create a MAC local address based on the process id.

    Assuming /proc/sys/kernel/pid_max requires at most 6 digits (32.768
    requires only 5 and is linux default), the last 3 bytes of the mac address
    can be mapped to pid digits (leaving all the a-f combinations free for
    other uses).
    """
    digits = []
    remain = pid = os.getpid()
    for _ in range(3):
        digit = int(remain % 100)
        digits.insert(0, digit)
        remain = int((remain - digit) / 100)
    # FIXME: May be the formatting could happen during conversion from unicode
    # so that a /mask/ of '0e:00:00' could be used to mean: gimme a MAC address
    # starting with. Or use 0e:00:00:P:I:D asking PID to be replaced
    # -- vila 2019-09-16
    prefix = conf.get('qemu.mac.prefix')
    mac_format = prefix + ':{:02d}' * 3
    mac_address = mac_format.format(*digits)
    logger.debug('{} got MAC {} for pid {}'.format(
        conf.get('vm.name'), mac_address, pid))
    return mac_address


register(options.Option('qemu.mac.address', default=mac_from_pid,
                        help_string='''\
To use in qemu.networks when a bridge is available on the host as:

  -net bridge,br={qemu.bridge} -net nic,macaddr={qemu.mac.address}

https://en.wikipedia.org/wiki/MAC_address documents local MAC addresses.
'''))
register(options.Option('qemu.mac.prefix', default='0e:00:00',
                        help_string='''\
This is used when creating a MAC address from the current process id.

The default value for qemu.mac.address uses this prefix.
'''))
register(options.ListOption('qemu.graphics', separator=' ',
                            default='-nographic',
                            help_string='''\
The options defining the qemu graphic setup.
'''))
register(TimeoutsOption('qemu.setup_ip.timeouts', default='0, 300, 60',
                        help_string='''\
When waiting for qemu to setup an IP address.'''))
register(TimeoutsOption('qemu.setup_ssh.timeouts', default='0, 120, 10',
                        help_string='''When waiting for qemu to setup ssh.'''))
register(TimeoutsOption('qemu.init.timeouts', default='10, 60, 10',
                        help_string='''When waiting for qemu to start.'''))
register(TimeoutsOption('qemu.cloud_init.timeouts', default='0, 240, 20',
                        help_string='''\
When waiting for cloud-init completion.'''))
register(TimeoutsOption('qemu.stop.timeouts', default='0, 60, 10',
                        help_string='''When waiting for qemu to stop.'''))
register(options.Option('qemu.pid',
                        default=None,
                        from_unicode=options.int_from_store,
                        help_string='''The qemu process id when running.'''))

######################################################################
# nova options
######################################################################
register(options.Option('nova.username', default_from_env=['OS_USERNAME'],
                        default=options.MANDATORY,
                        help_string='''The Open Stack user name.

This is generally set via OS_USERNAME, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.Option('nova.password', default_from_env=['OS_PASSWORD'],
                        default=options.MANDATORY,
                        help_string='''The Open Stack password.

This is generally set via OS_PASSWORD, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.Option('nova.region_name',
                        default_from_env=['OS_REGION_NAME'],
                        default=options.MANDATORY,
                        help_string='''The Open Stack region name.

This is generally set via OS_REGION_NAME, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.Option('nova.tenant_name',
                        default_from_env=['OS_TENANT_NAME'],
                        default=options.MANDATORY,
                        help_string='''The Open Stack tenant name.

This is generally set via OS_TENANT_NAME, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.Option('nova.auth_url', default_from_env=['OS_AUTH_URL'],
                        default=options.MANDATORY,
                        help_string='''The Open Stack keystone url.

This is generally set via OS_AUTH_URL, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.ListOption('nova.flavors', default=None,
                            help_string='''\
A list of flavors for all supported clouds.

The first known one is used.
'''))
register(options.Option('nova.image', default=options.MANDATORY,
                        help_string='''The glance image to boot from.'''))
register(options.Option('nova.net_id', default=None,
                        help_string='''The network id for the vm.'''))

register(options.Option('nova.boot_timeout', default='600',
                        from_unicode=options.float_from_store,
                        help_string='''\
Max time to boot a nova instance (in seconds).'''))
register(TimeoutsOption('nova.setup_ip.timeouts', default='0, 120, 10',
                        help_string='''\
 When waiting for nova to setup an IP.'''))
register(TimeoutsOption('nova.setup_ssh.timeouts', default='0, 120, 10',
                        help_string='''When waiting for nova to setup ssh.'''))
# FIXME: 600 is a more realistic value but canonistack can be really slow and
# 3600 has been observed. There should be a better way to declare default
# values on a per cloud basis -- vila 2015-12-21
register(options.Option('nova.cloud_init_timeout', default='3600',
                        from_unicode=options.float_from_store,
                        help_string='''\
Max time for cloud-init to finish (in seconds).'''))
register(options.ListOption('nova.setup.digest.options',
                            default='nova.username, nova.password,'
                            'nova.region_name, nova.tenant_name, nova.image,'
                            'nova.flavors',
                            help_string='''\
A list of nova related options that are used to define a vm.

The values for these options are hashed to produce 'vm.setup.digest'.
'''))

######################################################################
# aws ec2 options
######################################################################
register(options.Option('aws.key',
                        default=options.MANDATORY,
                        help_string='''The aws access key id.'''))
register(options.Option('aws.secret',
                        default=options.MANDATORY,
                        help_string='''The aws secret access key.'''))
register(options.Option('aws.region',
                        default=options.MANDATORY,
                        help_string='''The aws region name.'''))
register(options.ListOption('ec2.instance.types', default='t2.micro',
                            help_string='''\
A list of instance types to choose from.

The first known one is used.
'''))
register(options.Option('ec2.free_tier',
                        default=True,
                        from_unicode=options.bool_from_store,
                        help_string='''\
Only select free tier instance types.'''))
# The image regexp for a given distribution
ec2_image_re_registry = registries.Registry()

register(options.RegistryOption('ec2.distribution.images',
                                registry=ec2_image_re_registry,
                                help_string='''\
The image regexp to use to select an image for a given distribution.'''))
ec2_image_re_registry.register(
    'ubuntu',
    '^ubuntu/images/hvm-ssd/ubuntu-{vm.release}-.*-{vm.architecture}-server.*',
    'Ubuntu images AMI regexp.')
ec2_image_re_registry.register(
    'amazon',
    '^amzn{vm.release}-ami-ecs-hvm-.*-ebs',
    'Amazon Linux images AMI regexp.')
ec2_image_re_registry.register(
    'debian',
    '^debian-{vm.release}-{vm.architecture}-.*',
    'Debian images AMI regexp.')

register(options.Option('ec2.image', default=options.MANDATORY,
                        help_string='''\
The image regexp to select images.

If this is not set, {vm.distribution} is used to select the right regexp from
{ec2.distribution.images}.
'''))
register(options.Option('ec2.image.id', default=None,
                        help_string='''The AMI id to boot from.

This helps disambiguate {ec2.image} by providing a unique AMI id.
'''))
register(options.ListOption('ec2.image.tags', separator=' ',
                            default='',
                            help_string='''\
A flatten list of key, value pairs to set as tags on the created image.
'''))
register(options.ListOption('ec2.image.owners', default='',
                            help_string='''\
The image owners to filter image candidates.

A distribution specific default value is used if not set.
'''))
register(options.Option('ec2.instance.id',
                        default=None,
                        help_string='''The ec2 instance id.'''))
register(options.ListOption('ec2.instance.tags', separator=' ',
                            default='',
                            help_string='''\
A flatten list of key, value pairs to set as tags on the created instance.
'''))
register(options.Option('ec2.published.id',
                        default=None,
                        help_string='''The published image id.
This is available as long as the instance is not terminated and is required
to unpublish a previously published image.
'''))
register(options.Option('ec2.subnet', default=options.MANDATORY,
                        help_string='''\
The subnet id for the vm.'''))
register(options.ListOption('ec2.security.groups', default=options.MANDATORY,
                            help_string='''\
The security groups for the vm.'''))
register(TimeoutsOption('ec2.boot.timeouts', default='10, 900, 60',
                        help_string='''\
When waiting for ec2 to boot.'''))
register(TimeoutsOption('ec2.setup_ip.timeouts', default='10, 240, 20',
                        help_string='''\
When waiting for ec2 to setup an IP.'''))
register(TimeoutsOption('ec2.setup_ssh.timeouts', default='0, 120, 10',
                        help_string='''When waiting for ec2 to setup ssh.'''))
register(TimeoutsOption('ec2.cloud_init.timeouts', default='0, 120, 20',
                        help_string='''\
When waiting for cloud-init completion.
'''))
register(TimeoutsOption('ec2.stop.timeouts', default='0, 240, 20',
                        help_string='''When waiting for ec2 to stop.'''))
register(TimeoutsOption('ec2.terminate.timeouts', default='0, 240, 20',
                        help_string='''When waiting for ec2 to terminate.'''))
register(TimeoutsOption('ec2.create_image.timeouts', default='0, 1200, 120',
                        help_string='''\
When waiting for ec2 to create an image.'''))
register(options.ListOption('ec2.setup.digest.options',
                            default='aws.key, aws.secret,'
                            'aws.region, ec2.type, ec2.image, ec2.image.id,'
                            'ec2.subnet',
                            help_string='''\
A list of ec2 related options that are used to define a vm.

The values for these options are hashed to produce 'vm.setup.digest'.
'''))
######################################################################
# lxd options
######################################################################

# FIXME: Using None as a default value fails
# byov.tests.test_lxd.TestInit.test_unknown_image. This needs
# investigation. Using [] as the default value works around the issue. This is
# a byoc issue and could be addressed by relaxing the constraint around
# unknown options (currently failing) that could be left as-is. The fallouts
# are unclear though.  -- vila 2016-07-12
register(options.ListOption('lxd.profiles', default=[],
                            help_string='''\
The profile names to apply to the container.
If empty, lxd will use the default one.

Beware of not using cloud-init in the profiles as this would conflict with
byov itself.
'''))
register(options.Option('lxd.image',
                        default='{{vm.distribution}.lxd.image}',
                        help_string='''The lxd image to boot from.'''))
register(options.Option('lxd.mac.address',
                        default=None,
                        help_string='''The MAC address to use.
This is intended to support stable IP addresses. Use with care.'''))
register(options.Option('lxd.remote',
                        default=None,
                        help_string='''The lxd remote server to use.'''))
# FIXME: Arguably the lxd.config namespace should automatically set all the
# corresponding lxd config options -- vila 2018-04-24
register(options.Option('lxd.config.boot.autostart',
                        default=False,
                        from_unicode=options.bool_from_store,
                        help_string='''\
Should the container be started at host boot.
'''))
register(TimeoutsOption('lxd.setup_ip.timeouts', default='0, 360, 90',
                        help_string='''\
When waiting for lxd to setup an IP.'''))
register(TimeoutsOption('lxd.setup_ssh.timeouts', default='0, 120, 60',
                        help_string='''When waiting for lxd to setup ssh.'''))
register(TimeoutsOption('lxd.cloud_init.timeouts', default='0, 120, 10',
                        help_string='''\
When waiting for cloud-init completion.'''))
register(LxdUserMountsOption('lxd.user_mounts', default=None,
                             help_string='''\
A list of '<host path>:<vm path>' to be mounted.

'host' paths can start with '~' which will be expanded to the user home
directory.
Relative 'host' paths are from the current directory.

'vm' paths must be absolute.

The host paths will be available at the vm paths with the user uid/gid access
rights.

Note: This requires /etc/subuid and /etc/subgid to contain a 'root:<id>:1' line
with id being the correct user (and group) id (i.e. the user running the
'byovm setup' command).
'''))
register(options.Option('lxd.user_mounts.host.uid',
                        default=os.getuid(),
                        help_string='''The host uid owning the mounts.'''))
register(options.Option('lxd.user_mounts.host.gid',
                        default=os.getuid(),
                        help_string='''The host gid owning the mounts.'''))
# '1000' is the default uid/gid in ubuntu and debian for the default vm user
register(options.Option('lxd.user_mounts.container.uid',
                        default='1000',
                        help_string='''The container uid using the mounts.'''))
register(options.Option('lxd.user_mounts.container.gid',
                        default='1000',
                        help_string='''The container gid using the mounts.'''))
register(options.Option('lxd.nesting', default=0, invalid='error',
                        from_unicode=options.int_from_store,
                        help_string='''\
How many level of nesting containers should be allowed for the vm.

This is used to configure the container for allowing nesting containers and
check that the /etc/subuid and /etc/subgid files provision enough ids for the
nested containers.

'0' means no nested containers.

Note that an unprivileged container can create nested privileged containers
which cannot compromise the host.
'''))

register(options.ListOption('lxd.setup.digest.options',
                            default='lxd.profiles, lxd.image, lxd.mac.address,'
                            'lxd.nesting, lxd.config.boot.autostart,'
                            'lxd.user_mounts,'
                            'lxd.user_mounts.host.uid,'
                            'lxd.user_mounts.host.gid,'
                            'lxd.user_mounts.container.uid,'
                            'lxd.user_mounts.container.gid,'
                            'lxd.remote',
                            help_string='''\
A list of lxd related options that are used to define a vm.

The values for these options are hashed to produce 'vm.setup.digest'.
'''))


######################################################################
# docker options
######################################################################
register(options.Option('docker.file',
                        default='./Dockerfile',
                        help_string='''\
The docker file to build the image.'''))
register(options.Option('docker.file.expanded',
                        default=None,
                        help_string='''\
The option expanded docker file to build the image.'''))
register(options.Option('docker.image',
                        default='{{vm.distribution}.docker.image}',
                        help_string='''The docker image name.'''))
register(options.Option('docker.container.id',
                        default=None,
                        help_string='''The docker container id.'''))
register(options.Option('docker.image.id',
                        default=None,
                        help_string='''The docker image id.'''))
register(options.Option('docker.registry.name',
                        default='test-registry',
                        help_string='''\
The docker image registry container name.'''))
register(options.Option('docker.registry.host',
                        default='localhost',
                        help_string='''The docker image registry host.'''))
register(options.Option('docker.registry.port',
                        default='5000',
                        from_unicode=options.int_from_store,
                        help_string='''The docker image registry port.'''))
register(options.ListOption(
    'docker.registry.command',
    default=('docker, run, -d, -p, {docker.registry.port}:5000,'
             ' --name, {docker.registry.name}, registry:2'),
    help_string='''docker command to run an image test registry.'''))
register(options.Option('docker.create.image.hook', default=None,
                        help_string='''\
A command that is executed *on the host* before building the image.

If options expansion is needed, '@script' will expand the options inside
'script' before execution.
'''))
register(options.ListOption(
    'docker.image.build.command',
    # Final '.' is the working directory
    default='docker, image, build, --file, {docker.file.expanded}, .',
    help_string='''docker command to build the container image.'''))
register(options.ListOption(
    'docker.container.create.command',
    default='docker, container, create',
    help_string='''docker command to create a container.'''))
register(options.ListOption(
    'docker.ports',
    default=None,
    help_string='''docker ports list as outside:inside pairs.'''))
register(DockerMountsOption(
    'docker.mounts', default=None,
    help_string='''\
A list of docker mount definitions (tuples using ':' as a separator)
- type must be either bind, tmpfs or volume.

See https://docs.docker.com/storage/bind-mounts/ for details.

Examples:
 bind:~/src:/src:readonly:slave
 tmpfs:/tmp:65536:1777  (size and chmod bits)
 volume:volname:/mnt
'''))
register(options.ListOption(
    'docker.container.start.command',
    default='docker, container, start, {vm.name}',
    help_string='''docker command to start the container.'''))
# FIXME: At least one use case requires -t but shell_captured works without it
# -- vila 2022-03-11
register(options.ListOption(
    'docker.container.shell.command',
    default='docker, container, exec, -it',
    help_string='''docker command to run a command in the container.'''))
register(options.ListOption(
    'docker.container.stop.command',
    default='docker, container, stop, {vm.name}',
    help_string='''docker command to stop the container.'''))
register(options.ListOption(
    'docker.image.tag.command',
    default=('docker, image, tag, {docker.image.id}, '
             ' {docker.registry.host}:{docker.registry.port}/{docker.image}'),
    help_string='''docker command to tag an image for publication.'''))
register(options.ListOption(
    'docker.image.publish.command',
    default=('docker, image, push, '
             '{docker.registry.host}:{docker.registry.port}/{docker.image}'),
    help_string='''docker command to publish an image.'''))
register(options.ListOption(
    'docker.image.unpublish.command',
    default=('docker, image, rm, -f,'
             '{docker.registry.host}:{docker.registry.port}/{docker.image}'),
    help_string='''docker command to unpublish an image.'''))
register(options.ListOption(
    'docker.container.teardown.command',
    default='docker, container, rm, {vm.name}',
    help_string='''docker command to teardown the container.'''))
register(options.ListOption(
    'docker.container.cp.command',
    default='docker, container, cp',
    help_string='''docker command to copy a file to/from a container.'''))

# FIXME: docker.setup.digest.options is missing -- vila 2022-03-09

######################################################################
# apt options
######################################################################

register(options.ListOption(
    'apt.command',
    default=('sudo, env, DEBIAN_FRONTEND=noninteractive,'
             ' apt-get, {apt.options}'),
    help_string='''apt-get command.'''))
register(options.ListOption('apt.options',
                            default='--option=Dpkg::Options::=--force-confold,'
                            ' --option=Dpkg::Options::=--force-unsafe-io,'
                            ' --assume-yes, --quiet, --no-install-recommends',
                            help_string='''apt-get install options.'''))
register(options.Option('apt.proxy', default=None, invalid='error',
                        help_string='''\
A local proxy for apt to avoid repeated .deb downloads.

Example:

  apt.proxy = http://192.168.0.42:3142
'''))
register(options.ListOption('apt.sources', default=None,
                            help_string='''\
A list of apt sources entries to be added to the default ones.

Cloud-init already setup /etc/apt/sources.list with appropriate entries. Only
additional entries need to be specified here.
'''))
register(options.ListOption('apt.update.timeouts',
                            default='15.0, 90.0, 240.0',
                            help_string='''apt-get update timeouts in seconds.

When apt-get update fails on hash sum mismatches, retry after the specified
timeouts. More values mean more retries.
'''))
register(options.ListOption('apt.upgrade.timeouts',
                            default='15.0, 90.0, 240.0',
                            help_string='''\
apt-get dist-upgrade timeouts in seconds.

When apt-get fails with retcode 100 (Could not get lock /var/lib/dpkg/lock),
retry after the specified timeouts. More values mean more retries.
'''))
register(options.ListOption('apt.setup.digest.options',
                            default='apt.options, apt.sources',
                            help_string='''\
A list of apt related options that are used to define a vm.

The values for these options are hashed to produce 'vm.setup.digest'.
'''))

# FIXME: Needs a way to disable apt-get dist-upgrade which is breaking for
# ubuntu trusty libssl -- vila 2016-09-10

######################################################################
# pip options
######################################################################
register(options.ListOption(
    'pip.command',
    default='pip3, {pip.options}',
    help_string='''pip command.'''))
register(options.ListOption('pip.options', default='',
                            help_string='''pip options.'''))
# FIXME: This probably needs a specific option type to support pip specifics
# (-r <file> at least) -- vila 2021-12-30
register(PackageListOption('pip.packages', default='',
                           help_string='''\
A list of package specifiers to be installed.

If the package name starts with '@', it is interpreted as a file name
and its content, one package by line, is inserted in the list.
pip packages are installed after {vm.setup_scripts} have been executed.
'''))
register(options.ListOption('pip.setup.digest.options',
                            default='pip.options, pip.packages',
                            help_string='''\
A list of pip related options that are used to define a vm.

The values for these options are hashed to produce 'vm.setup.digest'.
'''))

######################################################################
# dnf options
######################################################################
register(options.ListOption('dnf.command',
                            default='sudo, dnf, {dnf.options}',
                            help_string='''dnf command.'''))
register(options.ListOption('dnf.options', default='-y',
                            help_string='''dnf options.'''))
register(PackageListOption('dnf.packages', default='',
                           help_string='''\
A list of package specifiers to be installed.

If the package name starts with '@', it is interpreted as a file name
and its content, one package by line, is inserted in the list.
'''))
register(options.ListOption('dnf.upgrade.timeouts',
                            default='15.0',
                            help_string='''\
dnf upgrade timeouts in seconds.

When dnf fails, retry after the specified timeouts.
More values mean more retries.
'''))
register(options.ListOption('dnf.setup.digest.options',
                            default='pip.options',
                            help_string='''\
A list of pip related options that are used to define a vm.

The values for these options are hashed to produce 'vm.setup.digest'.
'''))

######################################################################
# ssh options
######################################################################
register(options.ListOption('ssh.server_keys', default=None,
                            help_string='''\
A list of paths to server ssh keys.

Both public and private keys can be provided. Accepted server ssh key types are
rsa, dsa and ecdsa. The file names should match <type>.*[.pub].
'''))
# FIXME: There is a tension between providing a default value and making sure
# the user understand the need for an ssh key used to connect to the vms and
# its public counterpart needed to be in ssh.authorized_keys for the vm to
# accept connections. Having a password-less key or relying on a user agent is
# another question that the user needs to answer as there is no good default
# value for that either. -- vila 2016-07-07. A passwordless (or not for that
# matter) dedicated key under ~/.config/byov sounds like the most elegant
# solution and can already be achieved by setting 'ssh.key' in the user config
# file -- vila 2016-07-22
register(options.PathOption('ssh.key', default='~/.ssh/id_rsa',
                            help_string='''\
The ssh key path used to access the guest.

'ssh.authorized_keys' is expected to contain the .pub companion.
'''))
register(options.ListOption('ssh.authorized_keys',
                            default='{ssh.key}.pub',
                            help_string='''\
A list of paths to public ssh keys to be authorized for {vm.user}.

The content of the paths will be added to the authorized_keys file for
{vm.user}.
'''))
register(options.Option('ssh.user', default='{vm.user}',
                        help_string='''\
Which user should be used for ssh access.
'''))
register(options.Option('ssh.host', default='{vm.ip}',
                        help_string='''\
Which host name or IP address should be used for ssh access.
'''))
register(options.Option('ssh.port', default=None,
                        help_string='''\
Which port should be used for ssh access.
'''))
register(options.ListOption('ssh.options',
                            default='-oUserKnownHostsFile=/dev/null,'
                            '-oStrictHostKeyChecking=no,'
                            '-oIdentitiesOnly=yes,'
                            '-oIdentityFile={ssh.key}',
                            help_string='''\
A list of ssh options to be used when connecting to the guest via ssh.'''))
register(options.ListOption('ssh.setup.digest.options',
                            default='ssh.key, ssh.options',
                            help_string='''\
A list of ssh related options that are used to define a vm.

The values for these options are hashed to produce 'vm.setup.digest'.
'''))


######################################################################
# logging options
######################################################################
def level_from_store(s):
    val = None
    try:
        # Yes, _levelNames and _nameToLevel are private, but better use a
        # private than duplicate its content and get out of date.
        if sys.version_info < (3,):
            valid_levels = logging._levelNames
        else:
            valid_levels = logging._nameToLevel
        val = valid_levels[s.upper()]
    except KeyError:
        pass
    return val


register(options.Option('logging.level', default='ERROR',
                        from_unicode=level_from_store,
                        default_from_env=['LOG_LEVEL'],
                        help_string='''\
Logging level (same as python: error, warning, info, debug).'''))

register(options.Option('logging.format',
                        default='%(asctime)s %(levelname)s %(message)s',
                        help_string='''\
Logging format (see python doc).'''))


######################################################################
# launchpad options
######################################################################
def default_lp_login(conf=None):
    lp_login_cmd = ['bzr', 'lp-login']
    try:
        ret, out, err = subprocesses.run(lp_login_cmd)
        return out.strip()
    # python2 doesn't provide FileNotFoundError. OSError is raised if bzr is
    # not available
    except (subprocesses.errors.CommandError, OSError):
        return getpass.getuser()


register(options.Option('launchpad.login', default=default_lp_login,
                        help_string='''\
The launchpad login used for launchpad ssh access from the guest.

This defaults (acqired from the host) to bzr lp-login or, if not set, to the
first environment variable set in LOGNAME, USER, LNAME, USERNAME.
'''))

######################################################################
# gitlab options
######################################################################


def default_gitlab_login(conf=None):
    gl_login_cmd = ['git', 'config', 'gitlab.login']
    try:
        ret, out, err = subprocesses.run(gl_login_cmd)
        return out.strip()
    # python2 doesn't provide FileNotFoundError. OSError is raised if git is
    # not available
    except (subprocesses.errors.CommandError, OSError):
        return getpass.getuser()


register(options.Option('gitlab.login', default=default_gitlab_login,
                        help_string='''\
The gitlab login used for ssh access from the guest.

This defaults (acquired from the host) to 'git config gitlab.login' or, if not
set, to the first environment variable set in LOGNAME, USER, LNAME, USERNAME.
'''))

######################################################################
# debian options
######################################################################
register(options.Option('debian.user',
                        default='debian',
                        help_string='''The default user for debian.'''))
register(options.Option('debian.user.shell',
                        default='/bin/bash',
                        help_string='''The default shell on debian.'''))
register(options.Option('debian.lxd.image',
                        default='debian/{vm.release}/{vm.architecture}',
                        help_string='''The debian lxd image to boot from.'''))
register(options.Option('debian.package.manager',
                        default='apt',
                        help_string='''The package manager for debian.'''))
# FIXME: debian.qemu.download.url debian.qemu.download.path ?
# https://cloud.debian.org/images/openstack/current/ ?
# https://cloud.debian.org/images/cloud/ ?
# -- vila 2019-11-14


######################################################################
# ubuntu options
######################################################################
register(options.Option('ubuntu.user',
                        default='ubuntu',
                        help_string='''The default user for ubuntu.'''))
register(options.Option('ubuntu.user.shell',
                        default='/bin/bash',
                        help_string='''The default shell on ubuntu.'''))
register(options.Option('ubuntu.docker.image',
                        default='ubuntu:{vm.release}',
                        help_string='''\
The ubuntu docker image to boot from.'''))
register(options.Option('ubuntu.lxd.image',
                        default='ubuntu:{vm.release}/{vm.architecture}',
                        help_string='''The ubuntu lxd image to boot from.'''))
register(options.Option('ubuntu.package.manager',
                        default='apt',
                        help_string='''The package manager for ubuntu.'''))
# FIXME: ubuntu.qemu.download.url ubuntu.qemu.download.path ?
# -- vila 2019-11-14
######################################################################
# amazon options
######################################################################
register(options.Option('amazon.user',
                        default='ec2-user',
                        help_string='''The default user for amazon linux.'''))
register(options.Option('amazon.user.shell',
                        default='/bin/bash',
                        help_string='''The default shell on amazon linux.'''))
register(options.Option('amazon.docker.image',
                        default='amazonlinux',
                        help_string='''\
The amazon docker image to boot from.'''))
register(options.Option('amazon.lxd.image',
                        default='amazonlinux/{vm.release}/{vm.architecture}',
                        help_string='''The amzon lxd image to boot from.'''))
register(options.Option('amazon.package.manager',
                        default='dnf',  # Not there yet...
                        help_string='''The package manager for amazon.'''))
######################################################################
# centos options
######################################################################
register(options.Option('centos.user',
                        default='centos',
                        help_string='''The default user for centos.'''))
register(options.Option('centos.user.shell',
                        default='/bin/bash',
                        help_string='''The default shell on centos.'''))
register(options.Option('centos.package.manager',
                        default='dnf',
                        help_string='''The package manager for centos.'''))
register(options.Option(
    'centos.lxd.image',
    default='images:centos/{vm.release}/cloud/{vm.architecture}',
    help_string='''The centos lxd image to boot from.'''))
######################################################################
# fedora options
######################################################################
register(options.Option('fedora.user',
                        default='fedora',
                        help_string='''The default user for fedora.'''))
register(options.Option('fedora.user.shell',
                        default='/bin/bash',
                        help_string='''The default shell on fedora.'''))
register(options.Option('fedora.package.manager',
                        default='dnf',
                        help_string='''The package manager for fedora.'''))
register(options.Option(
    'fedora.lxd.image',
    default='images:fedora/{vm.release}/cloud/{vm.architecture}',
    help_string='''The fedora lxd image to boot from.'''))



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
"""Utilities to define, access and process the various log files Vms produce.

Various log files are produced on the host but contains information about vms
internals.

Various tricks are currently involved as while the user want full control on
this file, they are often created or populated with root access rights for the
file themselves or the directories where they leave.

Some information in those files may be acquired in better ways than scrapping
log files. Yet, this is a useful fallback until this information is available
from other sources.
"""

import re

# Note: There may be a simpler way than defining classes to decorate python
# regexps but it's not obvious to get that right. For now, this will do.


class Regexp(object):

    def __init__(self, string, flags=0):
        """Regexp constructor.

        :param string: The regular expression textual representation.

        :params flags: The 're.compile' applicable flags.
        """
        self.string = string
        self.flags = flags
        self.re = None

    def ensure_re(self):
        if self.re is None:
            self.re = re.compile(self.string, self.flags)

    def match(self, string):
        self.ensure_re()
        return self.re.match(string)

# FIXME: The following regexp is searched for in cloud-init output. This design
# fails when cloud-init is disabled (see http://pad.lv/1669675)
# -- vila 2019-10-16

# The following lines have been observed for Ubuntu releases from precise to
# trusty (IP replaced with x.x.x.x to please pep8, /me sighs):

# precise:
# ci-info: eth0  : 1  x.x.x.x   255.255.255.0   00:16:3e:9d:7d:2a
# quantal:
# ci-info: |  eth0  | True | x.x.x.x | 255.255.255.0 | 00:16:3e:5d:83:2d |
# raring
# ci-info: |  eth0  | True | x.x.x.x | 255.255.255.0 | 00:16:3e:87:7b:38 |
# saucy:
# ci-info: |  eth0  | True | x.x.x.x | 255.255.255.0 | 00:16:3e:5f:5d:48 |
# trusty:
# ci-info: |  eth0  | True | x.x.x.x | 255.255.255.0 | 00:16:3e:38:8d:7e |
# wily/systemd appeared briefly bug get reverted:
# ci-info: |  ens3  | True | x.x.x.x | 255.255.255.0 |   .   | <MAC addr> |
# and ens3 is back with xenial
# ci-info: |  ens3  | True | x.x.x.x | 255.255.255.0 |   .   | <MAC addr> |


# Which can be captured with the following regexps:

ip_address_re = Regexp(r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}')
mac_address_re = Regexp(r'[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}:'
                        r'[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}')


# We use a template so the interface name can be injected at run time

class InterfaceRegexp(Regexp):

    def __init__(self, iface, flags=0):
        super(InterfaceRegexp, self).__init__(None, flags=flags)
        self.iface = iface

    def ensure_re(self):
        if self.re is None:
            # We could generalize to a TemplatedRegexp accepting a template and
            # some format parameters (and flags ?) but until we need it, it's
            # overkill.

            # FIXME: vivid seems to say it's not overkill anymore...  the
            # output describing the interface definition has changed. Tests
            # missing. -- vila 2015-05-10
            template = r'.*ci-info: .*{0}.*?({1}).*?({1}).*?({2})'
            self.string = template.format(self.iface,
                                          ip_address_re.string,
                                          mac_address_re.string)
            self.re = re.compile(self.string, self.flags)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2022 Vincent Ladeuil.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
"""The docker backend is significantly different from other ones as it doesn't
rely on cloud-init at all because most features are already supported in docker
files.

Similarly, because docker provides an easy shell access, ssh is not needed nor
used.

Instead, we rely on docker files expanding them from the byov provided
configuration files.

"""
import json
import hashlib
import logging
import os
import pty
import tempfile

from byov import (
    errors,
    subprocesses,
    vms,
)


logger = logging.getLogger(__name__)


def container_info(vm_name):
    info = dict(state='UNKNOWN')
    cmd = ['docker', 'container', 'inspect', vm_name]
    ret, out, err = subprocesses.run(cmd, raise_on_error=False)
    jout = json.loads(out)
    if jout:
        # We ask for a single container inspection
        jout = jout[0]
        # Extract only the info we need
        # Finally, translate the state
        # https://stackoverflow.com/questions/32427684/
        # what-are-the-possible-states-for-a-docker-container#32428199
        docker_states = dict(created='STOPPED',
                             restarting='RUNNING',
                             running='RUNNING',
                             paused='STOPPED',
                             exited='STOPPED',
                             dead='UNKNOWN')
        info['state'] = docker_states[jout['State']['Status']]
        pass
    return info


class Docker(vms.VM):

    setup_ip_timeouts = 'docker.setup_ip.timeouts'
    existing_vm_options = vms.VM.existing_vm_options + [
        'docker.image.id', 'docker.container.id']

    def state(self):
        # man lxc(7) defines the possible states as: STOPPED, STARTING,
        # RUNNING, ABORTING, STOPPING. We add UNKNOWN.
        info = container_info(self.conf.get('vm.name'))
        return info['state']

    def hash_setup(self):
        hasher = hashlib.md5()
        # Seed with the digest from the base class to capture common options
        up = super().hash_setup().encode('utf8')
        hasher.update(up)
        options = self.conf.get('docker.setup.digest.options')
        if options:
            for opt in options:
                value = self.conf.get(opt, convert=False)
                self.hash_value(hasher, value)
        return hasher.hexdigest()

    def run_create_image_hook(self):
        """Locally run a setup command defined by the user."""
        return self.run_hook('docker.create.image.hook')

    def create_image(self):
        logger.info('Creating {} docker image...'.format(
            self.conf.get('docker.image')))
        self.run_create_image_hook()
        with tempfile.NamedTemporaryFile(mode='w', encoding='utf8',
                                         dir=os.getcwd()) as f:
            content = self.expand_file(self.conf.get('docker.file'))
            f.write(content)
            f.flush()
            self.conf.set('docker.file.expanded', f.name)
            create_command = self.conf.get('docker.image.build.command')
            ret, out, err = subprocesses.run(create_command)
            if '-q' in create_command:
                # last line of the output has the produced image id
                # drop eol and get image id from
                # 'sha256:c059bfaa......849c4d8e\n'
                img_id = out[0:-1].split(':')[1]
            else:
                img_id = out.splitlines()[-1][len('Successfully built '):]
            self.econf.set('docker.image.id', img_id)
            self.save_existing_config()

    def create(self):
        logger.info('Creating {} docker container...'.format(
            self.conf.get('vm.name')))
        create_command = self.conf.get('docker.container.create.command')
        create_command.extend(['--name', self.conf.get('vm.name')])
        mounts = self.conf.get('docker.mounts')
        if mounts:
            create_command.extend(mounts)
        ports = self.conf.get('docker.ports')
        if ports:
            create_command.extend(['-p=' + p for p in ports])

        create_command.append(self.conf.get('docker.image.id'))
        ret, out, err = subprocesses.run(create_command)
        self.save_existing_config()

    def start(self):
        super().start()
        logger.info('Starting {} docker container...'.format(
            self.conf.get('vm.name')))
        start_command = self.conf.get('docker.container.start.command')
        ret, out, err = subprocesses.run(start_command)
        self.save_existing_config()

    def upload(self, local_path, remote_path, mode_bits=None):
        logger.info('Copying {} into {} docker container...'.format(
            local_path, self.conf.get('vm.name')))
        if mode_bits is None:
            # Keep only the last 3 digits as that's what chmod will
            # accept. This will raises if path doesn't exist
            mode_bits = oct(self._mode_bits(local_path))[-3:]
        cp_command = self.conf.get('docker.container.cp.command')
        cp_command += [local_path]
        cp_command += ['{}:{}'.format(self.conf.get('vm.name'), remote_path)]
        ret, out, err = subprocesses.run(cp_command)
        chmod_command = self.conf.get('docker.container.shell.command')
        chmod_command += [self.conf.get('vm.name'),
                          'chmod', mode_bits, remote_path]
        ret, out, err = subprocesses.run(chmod_command)

    def shell(self, command, *args):
        if command and command.startswith('@'):
            retcode, out, err = self.run_script(
                command[1:], args=args, captured=False)
        else:
            # docker fails to properly connect if it doesn't get a pty
            cmd = self.conf.get('docker.container.shell.command')
            cmd += [self.conf.get('vm.name')]
            if not command:
                cmd += [self.conf.get('vm.user.shell')]
            else:
                cmd += [command] + list(args)
            retcode = pty.spawn(cmd)
            out, err = None, None  # nope, can't get that easily from pty
            if retcode:
                raise errors.CommandError(cmd, retcode, out, err)
        return retcode, out, err

    def get_ssh_command(self, command, *args):
        """Just a root shell access."""
        cmd = self.conf.get('docker.container.shell.command')
        cmd += [self.conf.get('vm.name')]
        if command is not None:
            # There is no sudo (and it's not needed, we *are* root)
            if command != 'sudo':
                cmd += [command]
            if args:
                cmd += args
        return cmd

    def setup(self):
        # https://stackoverflow.com/questions/37744961/docker-run-vs-create
        logger.info('Setting up {} docker container...'.format(
            self.conf.get('vm.name')))
        self.create_image()
        self.create()
        self.start()
        self.setup_over_ssh()

    def stop(self):
        logger.info('Stopping {} docker container...'.format(
            self.conf.get('vm.name')))
        stop_command = self.conf.get('docker.container.stop.command')
        ret, out, err = subprocesses.run(stop_command)

    def publish(self):
        img_name = self.conf.get('vm.published_as')
        logger.info('Publishing {} docker image from {}...'.format(
            img_name, self.conf.get('vm.name')))
        tag_command = self.conf.get('docker.image.tag.command')
        ret, out, err = subprocesses.run(tag_command)
        publish_command = self.conf.get('docker.image.publish.command')
        ret, out, err = subprocesses.run(publish_command)

    def unpublish(self):
        img_name = self.conf.get('vm.published_as')
        logger.info('Un-publishing {} docker image from {}...'.format(
            img_name, self.conf.get('vm.name')))
        unpublish_command = self.conf.get('docker.image.unpublish.command')
        ret, out, err = subprocesses.run(unpublish_command)

    def teardown(self, force=False):
        logger.info('Tearing down {} docker container...'.format(
            self.conf.get('vm.name')))
        if force and self.state() == 'RUNNING':
            self.stop()
        teardown_command = self.conf.get('docker.container.teardown.command')
        ret, out, err = subprocesses.run(teardown_command)
        # FIXME: delete the image too ? -- vila 2022-01-14
        super().teardown()



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2019 Vincent Ladeuil.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import contextlib
import errno
import io
import logging
import os
import socket
import subprocess
import time
import sys
import select


import pexpect


from pexpect import fdpexpect


from byoc import (
    registries,
)
from byov import (
    config,
    errors,
    subprocesses,
    vms,
)


logger = logging.getLogger(__name__)


class ImageCommand(object):

    def __init__(self, vm):
        self.vm = vm
        self.conf = vm.conf
        # Some defaults shared by most implementations
        self.target = self.conf.get('qemu.image')
        self.teardown_cmd = ['rm', '-f', '{target}']

    def run_cmd(self, cmd):
        expanded = [self.conf.expand_options(c, dict(source=self.source,
                                                     target=self.target))
                    for c in cmd]
        logger.info('Running {}'.format(' '.join(expanded)))
        return subprocesses.run(expanded)

    def setup(self):
        self.vm.ensure_dir(os.path.dirname(self.target))
        return self.run_cmd(self.setup_cmd)

    def teardown(self):
        return self.run_cmd(self.teardown_cmd)


class ImageClone(ImageCommand):
    '''Clone an existing image.'''

    def __init__(self, conf):
        super().__init__(conf)
        self.setup_cmd = self.conf.get('qemu.clone', expand=False)
        self.source = self.vm.backing_image_path()


class ImageConvert(ImageCommand):
    '''Convert an existing image.'''

    def __init__(self, conf):
        super().__init__(conf)
        self.setup_cmd = self.conf.get('qemu.convert', expand=False)
        self.source = self.conf.get('qemu.download.path')
        self.setup_cmd = ['qemu-img', 'convert', '-O', 'qcow2',
                          '{source}', '{target}']


class ImageCopy(ImageCommand):
    '''Copy an existing image.'''

    def __init__(self, conf):
        super().__init__(conf)
        self.source = self.conf.get('qemu.copy.input', expand=False)
        self.setup_cmd = ['cp', '{source}', '{target}']


class ImageCreate(ImageCommand):
    '''Create an image.'''

    def __init__(self, conf):
        super().__init__(conf)
        self.target = self.conf.get('qemu.image', expand=False)
        self.setup_cmd = ['qemu-img', 'create', '-f', 'qcow2',
                          '{target}']


class ImageDownload(ImageCommand):
    '''Download an image.'''

    def __init__(self, conf):
        super().__init__(conf)
        self.setup_cmd = self.conf.get('qemu.download', expand=False)
        self.source = self.conf.get('qemu.download.url')
        self.target = self.conf.get('qemu.download.path')

    def setup(self):
        logger.info('Downloading {}'.format(self.source))
        if not os.path.exists(self.target):
            super().setup()
        else:
            logger.info('Already at {}'.format(self.target))


class ImageResize(ImageCommand):
    '''Resize an image.'''

    def __init__(self, conf):
        super().__init__(conf)
        self.setup_cmd = self.conf.get('qemu.resize', expand=False)
        disk_size = self.conf.get('vm.disk_size')
        self.source = disk_size
        self.target = self.conf.get('qemu.image')
        if not disk_size:
            self.setup_cmd = ['/bin/true']
        self.teardown_cmd = ['/bin/true']

    def setup(self):
        if self.source:
            return self.run_cmd(self.setup_cmd)


# The image related commands
image_cmd_registry = registries.Registry()


def imgcmd_register(name, kls):
    image_cmd_registry.register(name, kls, kls.__doc__)

imgcmd_register('clone', ImageClone)
imgcmd_register('convert', ImageConvert)
imgcmd_register('copy', ImageCopy)
imgcmd_register('create', ImageCreate)
imgcmd_register('download', ImageDownload)
imgcmd_register('resize', ImageResize)


# FIXME: Require newer version for pexpect (>~=4.6)? -- vila 2019-12-09
def select_ignore_interrupts(iwtd, owtd, ewtd, timeout=None):

    '''This is a wrapper around select.select() that ignores signals. If
    select.select raises a select.error exception and errno is an EINTR
    error then it is ignored. Mainly this is used to ignore sigwinch
    (terminal resize). '''

    # if select() is interrupted by a signal (errno==EINTR) then
    # we loop back and enter the select() again.
    if timeout is not None:
        end_time = time.time() + timeout
    while True:
        try:
            return select.select(iwtd, owtd, ewtd, timeout)
        except InterruptedError:
            err = sys.exc_info()[1]
            if err.args[0] == errno.EINTR:
                # if we loop back we have to subtract the
                # amount of time we already waited.
                if timeout is not None:
                    timeout = end_time - time.time()
                    if timeout < 0:
                        return([], [], [])
            else:
                # something else caused the select.error, so
                # this actually is an exception.
                raise


# FIXME: upstream wants 'fd' to support os.fstat(fd) which our socket doesn't
# support. So we get rid of the restriction as the code works fine with a
# socket.
# -- vila 2019-12-09
class fdspawn(fdpexpect.fdspawn):

    def __init__(self, fd, args=None, timeout=30, maxread=2000,
                 searchwindowsize=None, logfile=None, encoding=None,
                 codec_errors='strict', use_poll=False):

        if hasattr(fd, 'fileno'):
            fd = fd.fileno()

        self.args = None
        self.command = None
        fdpexpect.SpawnBase.__init__(self, timeout, maxread, searchwindowsize,
                                     logfile, encoding=encoding,
                                     codec_errors=codec_errors)
        self.child_fd = fd
        self.own_fd = False
        self.closed = False
        self.name = '<file descriptor %d>' % fd
        self.use_poll = use_poll

    def read_nonblocking(self, size=1, timeout=-1):
        """
        Read from the file descriptor and return the result as a string.

        The read_nonblocking method of :class:`SpawnBase` assumes that a call
        to os.read will not block (timeout parameter is ignored). This is not
        the case for POSIX file-like objects such as sockets and serial ports.

        Use :func:`select.select`, timeout is implemented conditionally for
        POSIX systems.

        :param int size: Read at most *size* bytes.
        :param int timeout: Wait timeout seconds for file descriptor to be
            ready to read. When -1 (default), use self.timeout. When 0, poll.
        :return: String containing the bytes read
        """
        if os.name == 'posix':
            if timeout == -1:
                timeout = self.timeout
            rlist = [self.child_fd]
            wlist = []
            xlist = []
            rlist, wlist, xlist = select_ignore_interrupts(
                rlist, wlist, xlist, timeout)
            if self.child_fd not in rlist:
                raise pexpect.TIMEOUT('Timeout exceeded.')
        return super(fdspawn, self).read_nonblocking(size)


@contextlib.contextmanager
def MonitorSocket(qemu):
    try:
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        monitor_path = qemu.monitor_path()
        logger.debug('Connecting to {}...'.format(monitor_path))
        sock.connect(monitor_path)
        yield sock
    finally:
        try:
            sock.close()
        except OSError as e:
            if e.args[0] == 9:
                # bad file descriptor, qemu closed the socket
                pass
            else:
                raise


@contextlib.contextmanager
def Monitor(qemu):
    with MonitorSocket(qemu) as sock:
        monitor = None
        try:
            monitor = fdspawn(sock, timeout=10)
            yield monitor
        finally:
            if monitor is not None:
                try:
                    monitor.close()
                except OSError as e:
                    if e.args[0] == 9:  # Bad file descriptor, already closed
                        pass
                    raise


class Qemu(vms.VM):
    """QEMU implementation.

    A qemu process is launched in the background with a monitor socket and log
    files for stdout/stderr.

    All interactions happen via the monitor to get the state, start, stop or
    shut down the vm.
    """

    setup_ssh_timeouts = 'qemu.setup_ssh.timeouts'

    existing_vm_options = vms.VM.existing_vm_options + [
        'qemu.bridge',
        'qemu.download.dir',
        'qemu.download.path',
        'qemu.download.url',
        'qemu.graphics',
        'qemu.mac.address',
        'qemu.networks',
        'qemu.pid',  # FIXME: What happens on host reboot ? -- vila 2021-11-30
        'qemu.ssh.localhost.port',
        'qemu.image',
        'qemu.images.dir',
        'qemu.mac.address',
        'qemu.mac.prefix',
    ]

    def __init__(self, conf):
        super(Qemu, self).__init__(conf)
        # FIXME: cleanup _seed_path and friends, they
        # aren't really needed -- vila 2019-12-07
        # Disk paths
        self._seed_path = None

    def state(self):
        logger.debug('Find {} qemu state'.format(self.conf.get('vm.name')))
        state = 'UNKNOWN'
        # FIXME: https://qemu.weilnetz.de/doc/qemu-doc.html#Network-options
        # mentions :
        # -pidfile file
        # Store the QEMU process PID in file. It is useful if you launch QEMU
        # from a script.
        # -- vila 2019-09-10
        pid = self.conf.get('qemu.pid')
        if pid is None:
            return 'UNKNOWN'
        # If there is a pid, check it is still alive
        try:
            # Using signal 0 as kill(2) won't send a signal in that case but
            # still checks the pid exist.
            os.kill(pid, 0)
        except OSError:
            return 'STOPPED'
        # The pid exists, qemu is still running
        try:
            with Monitor(self) as monitor:
                monitor.expect('\(qemu\) ')
                # See there between RUNNING and STOPPED (suspended ?). Until
                # then, assume qemu is running when the monitor expects a
                # command.  Could be 'info status' returning running or paused
                # according to
                # https://qemu.readthedocs.io/en/latest/system/monitor.html
                state = 'RUNNING'
        except (ConnectionRefusedError, ConnectionResetError):
            logger.debug('Monitor cannot be reached')
            state = 'STOPPED'
        return state

    def spawn_qemu(self, with_seed=True):
        logger.info('Spawning {} qemu'.format(self.conf.get('vm.name')))
        try:
            return self._spawn_qemu(with_seed)
        except errors.ByovError:
            with io.open(self.stderr_path(), encoding='utf8') as f:
                stderr = f.read()
            logger.exception(
                'Spawning qemu for {} failed, qemu.stderr: {}'.format(
                    self.conf.get('vm.name'), stderr))
            raise

    def _spawn_qemu(self, with_seed=True):
        self.ensure_dir(self.config_dir_path())
        qemu_cmd = ['qemu-system-x86_64', '-enable-kvm']

        nb_cpus = self.conf.get('vm.cpus')
        if nb_cpus:
            qemu_cmd.extend(['-smp', nb_cpus])
        ram_size = self.conf.get('vm.ram_size')
        if ram_size:
            qemu_cmd.extend(['-m', ram_size])
        graphics = self.conf.get('qemu.graphics')
        if graphics:
            qemu_cmd.extend(graphics)
        networks = self.conf.get('qemu.networks')
        if networks:
            qemu_cmd.extend(networks)
        di_path = self.disk_image_path()
        if os.path.exists(di_path):
            qemu_cmd.extend(['-drive', 'if=virtio,file={}'.format(di_path)])
        more_disks = self.conf.get('qemu.disks')
        if more_disks:
            qemu_cmd.extend(more_disks)
        if with_seed and self._seed_path:
            qemu_cmd.extend(['-cdrom', self._seed_path])

        qemu_cmd.extend(
            ['-monitor', 'unix:' + self.monitor_path() + ',server'])

        def utf8_open(path):
            return io.open(path, 'w', encoding='utf8')

        # The child will inherit the file handles and take care of them. The
        # parent can close them safely.
        stdout_path = self.stdout_path()
        stderr_path = self.stderr_path()
        with utf8_open(stdout_path) as out, utf8_open(stderr_path) as err:
            logger.debug('Running {}'.format(' '.join(qemu_cmd)))
            proc = subprocess.Popen(qemu_cmd,
                                    stdin=subprocess.DEVNULL,
                                    stdout=out, stderr=err,
                                    close_fds=True)
        self.econf.set('qemu.pid', '{}'.format(proc.pid))
        self.save_existing_config()
        # At that point qemu is waiting for a connection to the monitor socket
        # This avoids a race between starting qemu up to the point where it
        # starts booting and monitoring it.
        qtos = self.conf.get('qemu.init.timeouts')
        me = self.spawn_qemu.__name__
        exc = None
        for attempt, sleep in enumerate(qtos, start=1):
            try:
                with MonitorSocket(self):
                    # Opening the socket is enough
                    logger.debug('qemu process {} is up'.format(proc.pid))
                    exc = None
                break  # We're done
            except ConnectionRefusedError as e:
                exc = e
                # FIXME: metric  -- vila 2019-09-11
                logger.debug(
                    'Sleeping {:.2f} seconds for {}/{} {}/{}'.format(
                        sleep, me, 'ConnectionRefusedError',
                        attempt, qtos.retries))
                time.sleep(sleep)
                continue  # Try again maybe ?
            except IOError as e:
                # The socket is not available yet
                # python2 does not provide FileNotFoundError
                if e.errno == errno.ENOENT:
                    # FIXME: metric  -- vila 2019-09-11
                    logger.debug(
                        'Sleeping {:.2f} seconds for {}/{} {}/{}'.format(
                            sleep, me, 'IOError', attempt, qtos.retries))
                    time.sleep(sleep)
                    continue  # Try again maybe ?
                else:
                    exc = e  # Try again maybe ?
        if exc is not None:  # We didn't succeed
            raise exc  # Re-raise the last seen exception

        # FIXME: The current design is quite ugly :-/ May be this means the ssh
        # access configuration is not appropriate (at least using port
        # forwarding looks messy). -- vila 2019-09-13

        # Two different network configurations are ~supported at this point,
        # differing mainly by how ssh access is obtained: port forwarding works
        # even if no other network connectivity is provided, whereas all other
        # configurations are supposed to reveal an IP address controlled by a
        # provided mac address. But depending on specifics, the IP is found
        # from the arp cache (populated by a ping on name) or in the cloud-init
        # console output (assuming cloud-init has not been disabled). Right now
        # only the arp/ping implementation is supported and it requires a
        # user-provided bridge on the host.  -- vila 2019-10-17
        if self.conf.get('qemu.ssh.localhost.port') is None:
            self.capture_ip()
            if self.conf.get('vm.ip') is None:
                raise errors.ByovError('{name} never received an IP',
                                       name=self.conf.get('vm.name'))
        logger.debug('IP address is {}'.format(self.conf.get('vm.ip')))

    def terminate_qemu(self):
        logger.info('Terminating {} qemu'.format(self.conf.get('vm.name')))
        try:
            with Monitor(self) as monitor:
                monitor.sendline('quit')
                monitor.expect(pexpect.EOF)
        except ConnectionRefusedError:
            # It's dead Jim
            logger.debug('Monitor cannot be reached', exc_info=True)

    def monitor_path(self):
        return os.path.join(self.config_dir_path(), 'monitor')

    def stdout_path(self):
        return os.path.join(self.config_dir_path(), 'qemu.stdout')

    def stderr_path(self):
        return os.path.join(self.config_dir_path(), 'qemu.stderr')

    def disk_image_path(self):
        return self.conf.get('qemu.image')

    def seed_path(self):
        return self.conf.get('qemu.image') + '.seed'

    def backing_image_path(self):
        backing_conf = config.VmStack(self.conf.get('vm.backing'))
        image_path = backing_conf.get('vm.published_as')
        return image_path

    def public_image_path(self):
        return self.conf.get('vm.published_as')

    def ips_from_arp(self, known=None):
        """Returns IP addresses for a given MAC address.

        This relies on arp and can be fed with a set of IPs known have been
        used in the past. This makes it possible to detect only (one) new IP
        when it's revealed.

        :param known: A list of known IPs as returned from a previous call.
        """
        if known is None:
            known = []
        mac_addr = self.conf.get('qemu.mac.address')
        bridge = self.conf.get('qemu.bridge')
        arp_cmd = ['arp', '-a', '-n', '-i', bridge]
        logger.debug('Running {}'.format(' '.join(arp_cmd)))
        arp = pexpect.spawn(arp_cmd[0], arp_cmd[1:])
        # Should match:
        # ? (192.168.0.178) at 0e:00:00:01:91:50 [ether] on br0
        ip_re = '\((\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\) at ' + mac_addr
        ips = []
        while True:
            ip_match = arp.expect([pexpect.EOF, ip_re])
            if ip_match:
                ip = arp.match.groups()[0].decode('utf8')
                if ip not in known:
                    ips.append(ip)
            else:
                break
        logger.debug('Found [{}] in arp output'.format(','.join(ips)))
        return ips

    def capture_ip(self):
        """Capture the qemu IP address in vm.ip.

        This avoids relying on cloud-init to get the IP address as long as
        'ping host' is enough to populate the arp cache. The underlying
        assumption is that the localhost should be able to resolve the host to
        IP translation (which is generally true when the dhcp server also
        populate the dns server(s)).
        """
        mac_addr = self.conf.get('qemu.mac.address')
        ping_cmd = 'ping -c1 {}'.format(self.conf.get('vm.name'))
        cpos = self.conf.get('qemu.setup_ip.timeouts')
        me = self.capture_ip.__name__
        # See what IPs are known for that mac address
        previous = self.ips_from_arp()
        for attempt, sleep in enumerate(cpos, start=1):
            logger.debug('Running {}'.format(ping_cmd))
            # If a new mac address is seen, the dhcp server should provide a
            # new IP. ping'ing the vm name should make that IP appear in the
            # arp tables.
            ping_output, ret = pexpect.run(ping_cmd, withexitstatus=True)
            ping_output = ping_output.decode('utf8')
            logger.debug('Ping output: {}'.format(ping_output))
            if not ret:
                # ping succeeded so the host can be found from the localhost it
                # should come with a new IP not seen before
                new_ips = self.ips_from_arp(previous)
                logger.debug('New IPs from arp: [{}]'.format(
                    ','.join(new_ips)))
                the_ip = 'not known yet'
                if len(new_ips) > 1:
                    raise errors.ByovError('{name} {mac} has weird {ips}',
                                           name=self.conf.get('vm.name'),
                                           mac=mac_addr, ips=new_ips)
                elif len(new_ips) == 1:
                    the_ip = new_ips[0]
                    self.econf.set('vm.ip', new_ips[0])
                else:
                    # No new IP appeared, from the ones known before the ping
                    # (somehow the host has already talked to the guest before
                    # the ping or when the guest is restarted), take the one
                    # appearing in the ping output
                    for p in previous:
                        if p in ping_output:
                            the_ip = p
                            break
                self.econf.set('vm.ip', the_ip)
                logger.info('{} IP is {}'.format(self.conf.get('vm.name'),
                                                 self.conf.get('vm.ip')))
                break
            logger.debug('Sleeping {:.2f} seconds for {} {}/{}'.format(
                sleep, me, attempt, cpos.retries))
            time.sleep(sleep)
        else:
            raise errors.ByovError('{name} {mac} never revealed an IP',
                                   name=self.conf.get('vm.name'),
                                   mac=mac_addr)

    def create_seed_image(self):
        if self._meta_data_path is None:
            self.create_meta_data()
        if self._user_data_path is None:
            self.create_user_data()
        seed_path = self.seed_path()
        self.ensure_dir(os.path.dirname(seed_path))
        subprocesses.run(
            ['genisoimage', '-output', seed_path,
             # cloud-init relies on the volid to discover its data
             '-volid', 'cidata',
             '-joliet', '-rock', '-input-charset', 'default',
             '-graft-points',
             'user-data=%s' % (self._user_data_path,),
             'meta-data=%s' % (self._meta_data_path,),
             ])
        self._seed_path = seed_path

    def setup_image(self):
        cmds = self.conf.get('qemu.image.setup')
        if cmds:
            # FIXME: If values need to be passed from command to command,
            # imgcmd.target may be involved... -- vila 2019-12-10
            for c in cmds:
                img = image_cmd_registry.get(c)
                imgcmd = img(self)
                imgcmd.setup()

    def teardown_image(self):
        cmds = self.conf.get('qemu.image.teardown')
        if cmds:
            for c in reversed(cmds):
                img = image_cmd_registry.get(c)
                img(self).teardown()

    def get_ssh_user_host_port(self):
        c = self.conf
        qemu_port = c.get('qemu.ssh.localhost.port')
        if qemu_port:
            return (c.get('ssh.user'), 'localhost', qemu_port)
        else:
            return super().get_ssh_user_host_port()

    def wait_for_cloud_init(self):
        self._wait_for_cloud_init('qemu.cloud_init.timeouts')

    def setup(self):
        logger.info('Setting up {} qemu...'.format(self.conf.get('vm.name')))
        if self._seed_path is None:
            self.create_seed_image()
        self.setup_image()
        self.spawn_qemu()
        self.wait_for_ssh()
        self.wait_for_cloud_init()
        self.setup_over_ssh()

    def start(self):
        super(Qemu, self).start()
        logger.info('Starting {} qemu...'.format(self.conf.get('vm.name')))
        # FIXME: Due to cloud-init disabling itself if the seed is not mounted
        # (see
        # https://bugs.launchpad.net/ubuntu/+source/cloud-init/+bug/1669675 and
        # https://bugs.launchpad.net/cloud-init/+bugs?field.tag=dsid), force
        # the seed below. Otherwise, we fail to find the IP address which is
        # provided by cloud-init -- vila 2019-10-15
        self.spawn_qemu(with_seed=True)
        self.wait_for_ssh()
        self.save_existing_config()

    def stop(self):
        logger.info('Stopping {} qemu...'.format(self.conf.get('vm.name')))
        with Monitor(self) as monitor:
            logger.debug('Sending system_powerdown...')
            monitor.sendline('system_powerdown')
            ret = monitor.expect([pexpect.EOF, pexpect.TIMEOUT])
            if ret:
                # We got a timeout, something bad is going on, escalate
                monitor.sendline('quit')
                try:
                    monitor.expect([pexpect.EOF, pexpect.TIMEOUT])
                except Exception:
                    # This will help understand when and why this can fail
                    logger.debug('monitor.expect failed', exc_info=True)
                    raise
                # FIXME: Report stderr on timeout here ? Start warning above
                # about qemu potentially left in an unbootable env ?
                # -- vila 2019-12-09

        # When the qemu process finish, no monitor connection can succeed
        stmos = self.conf.get('qemu.stop.timeouts')
        me = self.stop.__name__
        for attempt, sleep in enumerate(stmos, start=1):
            try:
                with MonitorSocket(self):
                    logger.debug('Socket is still up.')
            except ConnectionRefusedError:
                logger.debug('Socket is down.')
                break
            logger.debug('Sleeping {:.2f} seconds for {} {}/{}'.format(
                sleep, me, attempt, stmos.retries))
            time.sleep(sleep)
        else:
            raise errors.ByovError('{name} never stopped',
                                   name=self.conf.get('vm.name'))

    def teardown(self, force=False):
        if force and self.state() == 'RUNNING':
            self.stop()
        if self.state() == 'RUNNING':
            self.terminate_qemu()
        self.teardown_image()
        self.remove(self.seed_path())
        self.remove(self.stdout_path())
        self.remove(self.stderr_path())
        self.remove(self.monitor_path())
        super(Qemu, self).teardown()

    def publish(self):
        # FIXME: Probably worth rewriting with ImageCommand -- vila 2019-12-08
        # FIXME: ImageCommand should also provide a way to publish an image
        # from a disk other than the booting one -- vila 2019-12-19
        image_path = self.public_image_path()
        logger.info('Publishing qemu image {} from {}...'.format(
            image_path, self.conf.get('vm.name')))
        publish_command = ['cp', self.disk_image_path(), image_path]
        subprocesses.run(publish_command)

    def unpublish(self):
        # FIXME: Should published_as be registered in econf so that unpublish
        # use that ? That would allow removing an old image when the name is
        # changed. -- vila 2019-11-16
        # FIXME: Probably worth rewriting with ImageCommand -- vila 2019-12-08
        image_path = self.public_image_path()
        logger.info('Un-publishing {} qemu image from {}...'.format(
            image_path, self.conf.get('vm.name')))
        unpublish_command = ['rm', '-f', image_path]
        subprocesses.run(unpublish_command)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import errno
import unittest

from byov import (
    config,
    errors,
)
from byov.tests import (
    features,
    fixtures,
)
from byov.vms import lxd


class TestPush(unittest.TestCase):

    def setUp(self):
        super(TestPush, self).setUp()
        fixtures.setup_tests_config(self)
        features.requires_existing_vm(self, 'byov-tests-lxd-debian')
        fixtures.set_uniq_vm_name(self)
        # Create a shared config
        conf = config.VmStack(None)
        conf.store._load_from_string('''
[{vm_name}]
vm.name = {vm_name}
vm.class = ephemeral-lxd
vm.backing = byov-tests-lxd-debian
'''.format(vm_name=self.vm_name))
        conf.store.save()
        conf.store.unload()
        self.vm = lxd.EphemeralLxd(config.VmStack(self.vm_name))
        self.addCleanup(self.vm.stop)
        self.vm.start()

    def test_simple(self):
        local = 'foo'
        with open(local, 'w') as f:
            f.write('quux')
        remote = '/home/{}/bar'.format(self.vm.conf.get('vm.user'))
        ret = self.vm.push(local, remote)
        self.assertEqual(0, ret)
        ret, _, _ = self.vm.shell_captured('test', '-f', remote)
        self.assertEqual(0, ret)

    def test_expanded(self):
        local = 'foo'
        with open(local, 'w') as f:
            f.write('{vm.name}')
        remote = '/home/{}/bar'.format(self.vm.conf.get('vm.user'))
        ret = self.vm.push('@' + local, remote)
        self.assertEqual(0, ret)
        ret, out, _ = self.vm.shell_captured('cat', remote)
        self.assertEqual(0, ret)
        self.assertEqual(self.vm.conf.get('vm.name'), out)

    def test_unknown_src(self):
        local = 'foo'
        remote = '/home/{}/bar'.format(self.vm.conf.get('vm.user'))
        with self.assertRaises(OSError) as cm:
            self.vm.push(local, remote)
        self.assertEqual(errno.ENOENT, cm.exception.errno)
        self.assertEqual(local, cm.exception.filename)

    def test_unknown_dst(self):
        local = 'foo'
        with open(local, 'w') as f:
            f.write('quux')
        remote = '/bar'
        with self.assertRaises(errors.CommandError):
            self.vm.push(local, remote)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2016, 2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
"""Map configuration data into a cloud-init user-data."""

from __future__ import unicode_literals
import errno
import os
import io
import yaml

from byov import (
    errors,
    ssh,
)


class CIUserData(object):
    """Maps configuration data into cloud-init user-data.

    This is a container for the data that will ultimately be encoded into a
    cloud-config-archive user-data file.
    """

    def __init__(self, conf):
        super(CIUserData, self).__init__()
        self.conf = conf
        # The objects we will populate before creating a yaml encoding as a
        # cloud-config-archive file
        self.cloud_config = {}
        self.root_hook = None
        self.user_hook = None
        self.launchpad_hook = None
        self.uploaded_scripts_hook = None

    def set(self, ud_name, conf_name=None, value=None):
        """Set a user-data option from its corresponding configuration one.

        :param ud_name: user-data key.

        :param conf_name: configuration key, If set to None, `value` should be
            provided.

        :param value: value to use if `conf_name` is None.
        """
        if value is None and conf_name is not None:
            value = self.conf.get(conf_name)
        if value:
            self.cloud_config[ud_name] = value

    def _file_content(self, path, option_name):
        full_path = os.path.expanduser(path)
        try:
            with open(full_path) as f:
                content = f.read()
        except IOError as e:
            if e.args[0] == errno.ENOENT:
                raise errors.ConfigPathNotFound(path, option_name)
            else:
                raise
        return content

    def _key_from_path(self, path, option_name):
        """Infer user-data key from file name."""
        ssh_type, kind = ssh.infos_from_path(path)
        if ssh_type is None:
            raise errors.ConfigValueError(option_name, path)
        return '%s_%s' % (ssh_type, kind)

    def set_ssh_keys(self):
        """Set the server ssh keys from a list of paths.

        Provided paths should respect some coding:

        - the base name should start with the ssh type of their key (rsa, dsa,
          ecdsa),

        - base names ending with '.pub' are for public keys, the others are for
          private keys.
        """
        key_paths = self.conf.get('ssh.server_keys')
        if key_paths:
            ssh_keys = {}
            for p in key_paths:
                key = self._key_from_path(p, 'ssh.server_keys')
                ssh_keys[key] = self._file_content(p, 'ssh.server_keys')
            self.set('ssh_keys', None, ssh_keys)

    def set_apt_sources(self):
        sources = self.conf.get('apt.sources')
        if sources:
            apt_sources = []
            for src in sources:
                # '|' should not appear in urls nor keys so it should be safe
                # to use it as a separator.
                parts = src.split('|')
                if len(parts) == 1:
                    apt_sources.append({'source': parts[0]})
                else:
                    # For PPAs, an additional GPG key should be imported in the
                    # guest.
                    apt_sources.append({'source': parts[0], 'keyid': parts[1]})
            self.cloud_config['apt_sources'] = apt_sources

    def append_bootcmd(self, cmd):
        cmds = self.cloud_config.get('bootcmd', [])
        cmds.append(cmd)
        self.cloud_config['bootcmd'] = cmds

    def append_runcmd(self, cmd):
        cmds = self.cloud_config.get('runcmd', [])
        cmds.append(cmd)
        self.cloud_config['runcmd'] = cmds

    def _hook_script_path(self, user, when):
        return '/tmp/{}-byov_{}'.format(user, when)

    def _hook_content(self, option_name, user, hook_path, mode='0700'):
        # FIXME: Add more tests towards properly creating a tree on the guest
        # from a tree on the host. There seems to be three kind of items worth
        # caring about here: file content (output path, owner, chmod), file
        # (input and output paths, owner, chmod) and directory (path, owner,
        # chmod). There are also some subtle traps involved about when files
        # are created across various vm generations (one vm creates a dir, a mv
        # on top of that one doesn't, but still creates a file in this dir,
        # without realizing it can fail in a fresh vm). -- vila 2013-03-10
        host_path = self.conf.get(option_name)
        if host_path is None:
            return None
        fcontent = self._file_content(host_path, option_name)
        # Expand options in the provided content so we report better errors
        expanded_content = self.conf.expand_options(fcontent)
        # The following will generate an additional newline at the end of the
        # script. I can't think of a case where it matters and it makes this
        # code more robust (and/or simpler) if the script/file *doesn't* end up
        # with a proper newline.
        # FIXME: This may be worth fixing if we provide a more generic way to
        # create a remote tree. -- vila 2013-03-10
        hook_content = '''#!/bin/sh -e
cat >{__guest_path} <<'EOSETUPVMUNIQUECONTENTDONTBREAKFORFUN'
{__fcontent}
EOSETUPVMUNIQUECONTENTDONTBREAKFORFUN
chown {__user}:{__user} {__guest_path}
chmod {__mode} {__guest_path}
'''
        return hook_content.format(__user=user, __fcontent=expanded_content,
                                   __mode=mode,
                                   __guest_path=hook_path)

    def set_boot_hook(self):
        # FIXME: Needs a test ensuring we execute as root -- vila 2013-03-07
        hook_path = self._hook_script_path('root', 'pre_install')
        content = self._hook_content('vm.root_script', 'root', hook_path)
        if content is not None:
            self.root_hook = content
            # FIXME: This will run on all boots and the script should also be
            # deleted once executed -- vila 2016-05-11
            self.append_bootcmd(hook_path)

    def set_user_hook(self):
        # FIXME: Needs a test ensuring we execute as default user
        # -- vila 2013-03-07
        hook_path = self._hook_script_path('root', 'post_install')
        host_path = self.conf.get('vm.user_script')
        if host_path is None:
            return
        fcontent = self._file_content(host_path, 'vm.user_script')
        if fcontent is not None:
            # Expand options in the provided content so we report better errors
            self.user_hook = self.conf.expand_options(fcontent)
            self.append_runcmd(hook_path)

    def set_uploaded_scripts(self):
        script_paths = self.conf.get('vm.uploaded_scripts')
        if not script_paths:
            return
        hook_path = self.conf.expand_options('~{vm.user}/byov_uploads')
        bindir = self.conf.get('vm.uploaded_scripts.guest_dir')
        out = io.StringIO()
        out.write('''#!/bin/sh -e
cat >{hook_path} <<'EOSETUPVMUNIQUECONTENTDONTBREAKFORFUN'
mkdir -p {bindir}
cd {bindir}
'''.format(**locals()))
        for path in script_paths:
            fcontent = self._file_content(path, 'vm.uploaded_scripts')
            expanded = self.conf.expand_options(fcontent)
            base = os.path.basename(path)
            # FIXME: ~Duplicated from _hook_content. -- vila 2012-03-15
            out.write('''cat >{base} <<'EOF{base}'
{expanded}
EOF{base}
chmod 0755 {base}
'''.format(**locals()))

        out.write('''EOSETUPVMUNIQUECONTENTDONTBREAKFORFUN
chown {user}:{user} {hook_path}
chmod 0700 {hook_path}
'''.format(user=self.conf.get('vm.user'), **locals()))
        self.uploaded_scripts_hook = out.getvalue()
        self.append_runcmd(['su', '-l', '-c',
                            hook_path, self.conf.get('vm.user')])

    def set_users_groups(self):
        # FIXME: There is a risk that we need to play catch-up with cloud-init
        # changing default values (or running into a bug because we don't
        # properly cover the same defaults). Ideally, there should be a way to
        # revert to cloud-init 'default' user. -- vila 2018-03-06
        user = {
            'name': self.conf.get('vm.user'),
            'system': self.conf.get('vm.user.system'),
        }
        shell = self.conf.get('vm.user.shell')
        if shell:
            user['shell'] = shell
        home = self.conf.get('vm.user.home')
        if home:
            user['homedir'] = home
        ssh_keys = self.conf.get('ssh.authorized_keys')
        if ssh_keys:
            paths = [self._file_content(path, 'ssh.authorized_keys')
                     for path in ssh_keys]
            user['ssh-authorized-keys'] = paths
        sudos = self.conf.get('vm.user.sudo')
        if sudos:
            user['sudo'] = sudos
        self.set('users', None, [user])

    def populate(self):
        # The commands executed during init stage
        self.set_boot_hook()
        # Common and non-configurable options
        # login from the console can help diagnose network issues. This can be
        # authorized by using dict(expire=False, list=['ubuntu:ubuntu']) below
        # FIXME: That was needed again to debug a network related issue so
        # deserve an option -- vila 2019-10-15
        # The rationale is:
        # When something goes wrong in the stack leading to a working ssh
        # connection, debug can be achieved by connecting via the console. But
        # for that to work, the default user needs a valid password rather than
        # the default policy being to allow only ssh access with an authorized
        # key.
        chpasswd = self.conf.get('vm.chpasswd')
        if chpasswd:
            self.set('chpasswd', None, dict(expire=False, list=chpasswd))
        else:
            self.set('chpasswd', None, dict(expire=True))
        # Configurable options
        self.set('manage_etc_hosts', 'vm.manage_etc_hosts')
        self.set('fqdn', 'vm.fqdn')
        msg = self.conf.get('vm.final_message')
        if msg is None:
            if self.conf.get('vm.release') == 'precise':
                # Curse cloud-init lack of compatibility
                msg = 'byov finished installing in $UPTIME seconds.'
            else:
                msg = 'byov finished installing in ${uptime} seconds.'
        self.set('final_message', None, msg)
        self.set_users_groups()
        self.set_ssh_keys()
        self.set('apt_proxy', 'apt.proxy')
        self.set('locale', 'vm.locale')
        self.set_apt_sources()
        # upload the scripts
        self.set_uploaded_scripts()
        # The commands executed before powering off
        self.set_user_hook()

    def add_boot_hook(self, parts, hook):
        """Add a cloud-boothook.

        Note that this hook is excuted immediately, run at every boot at an
        early stage where parts of the system may not be available (including
        the default user).

        :params parts: The existing parts of the cloud-config-archive.

        :params hook: The hook to add in textual form.
        """
        if hook is not None:
            parts.append({'content': '#cloud-boothook\n' + hook})

    def add_user_data_script(self, parts, hook):
        """Add a User-Data script.

        The received script is executed very late in the boot sequence
        (rc.local-like).

        :params parts: The existing parts of the cloud-config-archive.

        :params hook: The hook to add in textual form. It MUST start with a
            shebang line '#!'.
        """
        if hook is not None:
            parts.append({'content': hook})

    def dump(self):
        def yaml_dumps(obj):
            return yaml.safe_dump(obj)

        parts = [{'content':
                  '#cloud-config\n' + yaml_dumps(self.cloud_config)}]
        self.add_boot_hook(parts, self.root_hook)
        self.add_user_data_script(parts, self.launchpad_hook)
        self.add_user_data_script(parts, self.uploaded_scripts_hook)
        self.add_user_data_script(parts, self.user_hook)
        # Wrap the lot into a cloud config archive
        return '#cloud-config-archive\n' + yaml_dumps(parts)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014, 2015, 2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
import io
import unittest


from byot import fixtures
from byov import (
    errors,
    monitors,
)
from byov.tests import (
    fixtures as vms_fixtures,
)


class TestConsoleParsing(unittest.TestCase):

    def _scan_console_monitor(self, content):
        mon = monitors.ConsoleMonitor(io.BytesIO(content))
        lines = []
        for line in mon.scan():
            lines.append(line)
        return lines

    def test_fails_on_empty(self):
        with self.assertRaises(errors.ConsoleEOFError):
            self._scan_console_monitor(b'')

    def test_fail_on_knwon_cloud_init_errors(self):
        with self.assertRaises(errors.CloudInitError):
            self._scan_console_monitor(b'Failed loading yaml blob\n')
        with self.assertRaises(errors.CloudInitError):
            self._scan_console_monitor(
                b'Unhandled non-multipart userdata starting\n')
        with self.assertRaises(errors.CloudInitError):
            self._scan_console_monitor(
                b"failed to render string to stdout: cannot find 'uptime'\n")
        with self.assertRaises(errors.CloudInitError):
            self._scan_console_monitor(
                b"Failed loading of cloud config "
                b"'/var/lib/cloud/instance/cloud-config.txt'. "
                b"Continuing with empty config\n")

    def test_succeds_on_final_message(self):
        lines = self._scan_console_monitor(b'''
Lalala
I'm doing my work
It goes nicely
byovm finished installing in 1 seconds.
That was fast isn't it ?
 * Will now halt
[   33.204755] Power down.
''')
        # We stop as soon as we get the final message and ignore the rest
        self.assertEqual(' * Will now halt\n', lines[-1])


class TestConsoleParsingWithFile(unittest.TestCase):

    def setUp(self):
        super(TestConsoleParsingWithFile, self).setUp()
        vms_fixtures.isolate_from_disk(self)

    def _scan_file_monitor(self, content):
        with open('console', 'w') as f:
            f.write(content)
        mon = monitors.TailMonitor('console')
        for line in mon.scan():
            pass
        return mon.lines

    def test_succeeds_with_file(self):
        content = '''\
Yet another install
Going well
byovm finished installing in 0.5 seconds.
Wow, even faster !
 * Will now halt
Whatever, won't read that
'''
        lines = self._scan_file_monitor(content)
        expected_lines = content.splitlines(True)
        # Remove the last line that should not be seen
        expected_lines = expected_lines[:-1]
        self.assertEqual(expected_lines, lines)

    def xtest_fails_on_empty_file(self):
        # FIXME: We need some sort of timeout there... -- vila 2013-ish
        with self.assertRaises(errors.CommandError):
            self._scan_file_monitor('')

    def test_fail_on_knwon_cloud_init_errors_with_file(self):
        with self.assertRaises(errors.CloudInitError):
            self._scan_file_monitor('Failed loading yaml blob\n')
        with self.assertRaises(errors.CloudInitError):
            self._scan_file_monitor(
                'Unhandled non-multipart userdata starting\n')
        with self.assertRaises(errors.CloudInitError):
            self._scan_file_monitor(
                "failed to render string to stdout: cannot find 'uptime'\n")


class TestActualFileSize(unittest.TestCase):

    def setUp(self):
        super(TestActualFileSize, self).setUp()
        fixtures.set_uniq_cwd(self)

    def assertSize(self, expected, path):
        self.assertEqual(expected, monitors.actual_file_size(path))

    def test_empty_file(self):
        open('foo', 'w').close()
        self.assertSize(0, 'foo')

    def test_file_with_content(self):
        with open('foo', 'w') as f:
            f.write('bar')
        self.assertSize(3, 'foo')

    def test_unknown_file(self):
        self.assertSize(None, 'I-dont-exist')
